##
# Copyright 2017-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing TensorFlow, implemented as an easyblock

@author: Kenneth Hoste (HPC-UGent)
@author: Ake Sandgren (Umea University)
@author: Damian Alvarez (Forschungzentrum Juelich GmbH)
@author: Alexander Grund (TU Dresden)
"""
import glob
import os
import re
import stat
import tempfile

import easybuild.tools.environment as env
import easybuild.tools.toolchain as toolchain
from easybuild.easyblocks.generic.pythonpackage import PythonPackage, det_python_version
from easybuild.easyblocks.python import EXTS_FILTER_PYTHON_PACKAGES
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools import run, LooseVersion
from easybuild.tools.build_log import EasyBuildError, print_warning
from easybuild.tools.config import build_option, IGNORE
from easybuild.tools.filetools import adjust_permissions, apply_regex_substitutions, copy_file, mkdir, resolve_path
from easybuild.tools.filetools import is_readable, read_file, which, write_file, remove_file
from easybuild.tools.modules import get_software_root, get_software_version, get_software_libdir
from easybuild.tools.run import run_cmd
from easybuild.tools.systemtools import X86_64, get_cpu_architecture, get_os_name, get_os_version


CPU_DEVICE = 'cpu'
GPU_DEVICE = 'gpu'

# Wrapper for Intel(MPI) compilers, where required environment variables
# are hardcoded to make sure they are present;
# this is required because Bazel resets the environment in which
# compiler commands are executed...
INTEL_COMPILER_WRAPPER = """#!/bin/bash

export CPATH='%(cpath)s'

# Only relevant for Intel compilers.
export INTEL_LICENSE_FILE='%(intel_license_file)s'

# Only relevant for MPI compiler wrapper (mpiicc/mpicc etc),
# not for regular compiler.
export I_MPI_ROOT='%(intel_mpi_root)s'

# Exclude location of this wrapper from $PATH to avoid other potential
# wrappers calling this wrapper.
export PATH=$(echo $PATH | tr ':' '\n' | grep -v "^%(wrapper_dir)s$" | tr '\n' ':')

%(compiler_path)s "$@"
"""


def split_tf_libs_txt(valid_libs_txt):
    """Split the VALID_LIBS entry from the TF file into single names"""
    entries = valid_libs_txt.split(',')
    # Remove double quotes and whitespace
    result = [entry.strip().strip('"') for entry in entries]
    # Remove potentially trailing empty element due to trailing comma in the txt
    if not result[-1]:
        result.pop()
    return result


def get_system_libs_from_tf(source_dir):
    """Return the valid values for TF_SYSTEM_LIBS from the TensorFlow source directory"""
    syslibs_path = os.path.join(source_dir, 'third_party', 'systemlibs', 'syslibs_configure.bzl')
    result = []
    if os.path.exists(syslibs_path):
        txt = read_file(syslibs_path)
        valid_libs_match = re.search(r'VALID_LIBS\s*=\s*\[(.*?)\]', txt, re.DOTALL)
        if not valid_libs_match:
            raise EasyBuildError('VALID_LIBS definition not found in %s', syslibs_path)
        result = split_tf_libs_txt(valid_libs_match.group(1))
    return result


def get_system_libs_for_version(tf_version, as_valid_libs=False):
    """
    Determine valid values for $TF_SYSTEM_LIBS for the given TF version

    If as_valid_libs=False (default) then returns 2 dictioniaries:
        1: Mapping of <EB name> to <TF name>
        2: Mapping of <package name> to <TF name> (for python extensions)
    else returns a string formated like the VALID_LIBS variable in third_party/systemlibs/syslibs_configure.bzl
        Those can be used to check/diff against third_party/systemlibs/syslibs_configure.bzl by running:
            python -c 'from easybuild.easyblocks.tensorflow import get_system_libs_for_version; \
                        print(get_system_libs_for_version("2.1.0", as_valid_libs=True))'
    """
    tf_version = LooseVersion(tf_version)

    def is_version_ok(version_range):
        """Return True if the TF version to be installed matches the version_range"""
        min_version, max_version = version_range.split(':')
        result = True
        if min_version and tf_version < LooseVersion(min_version):
            result = False
        if max_version and tf_version >= LooseVersion(max_version):
            result = False
        return result

    # For these lists check third_party/systemlibs/syslibs_configure.bzl --> VALID_LIBS
    # Also verify third_party/systemlibs/<name>.BUILD or third_party/systemlibs/<name>/BUILD.system
    # if it does something "strange" (e.g. link hardcoded headers)

    # Software which is added as a dependency in the EC
    available_system_libs = {
        # Format: (<EB name>, <version range>): <TF name>
        #         <version range> is '<min version>:<exclusive max version>'
        ('Abseil', '2.9.0:'): 'com_google_absl',
        ('cURL', '2.0.0:'): 'curl',
        ('double-conversion', '2.0.0:'): 'double_conversion',
        ('flatbuffers', '2.0.0:'): 'flatbuffers',
        ('giflib', '2.0.0:2.1.0'): 'gif_archive',
        ('giflib', '2.1.0:'): 'gif',
        ('hwloc', '2.0.0:'): 'hwloc',
        ('ICU', '2.0.0:'): 'icu',
        ('JsonCpp', '2.0.0:'): 'jsoncpp_git',
        ('libjpeg-turbo', '2.0.0:2.2.0'): 'jpeg',
        ('libjpeg-turbo', '2.2.0:'): 'libjpeg_turbo',
        ('libpng', '2.0.0:2.1.0'): 'png_archive',
        ('libpng', '2.1.0:'): 'png',
        ('LMDB', '2.0.0:'): 'lmdb',
        ('NASM', '2.0.0:'): 'nasm',
        ('nsync', '2.0.0:'): 'nsync',
        ('PCRE', '2.0.0:2.6.0'): 'pcre',
        ('protobuf', '2.0.0:'): 'com_google_protobuf',
        ('pybind11', '2.2.0:'): 'pybind11',
        ('snappy', '2.0.0:'): 'snappy',
        ('SQLite', '2.0.0:'): 'org_sqlite',
        ('SWIG', '2.0.0:2.4.0'): 'swig',
        ('zlib', '2.0.0:2.2.0'): 'zlib_archive',
        ('zlib', '2.2.0:'): 'zlib',
    }
    # Software recognized by TF but which is always disabled (usually because no EC is known)
    # Format: <TF name>: <version range>
    unused_system_libs = {
        'boringssl': '2.0.0:',
        'com_github_googleapis_googleapis': '2.0.0:2.5.0',
        'com_github_googlecloudplatform_google_cloud_cpp': '2.0.0:',  # Not used due to $TF_NEED_GCP=0
        'com_github_grpc_grpc': '2.2.0:',
        'com_googlesource_code_re2': '2.0.0:',
        'grpc': '2.0.0:2.2.0',
    }
    # Python packages installed as extensions or in the Python module
    # Will be checked for availabilitly
    # Format: (<package name>, <version range>): <TF name>
    python_system_libs = {
        ('absl', '2.0.0:'): 'absl_py',
        ('astor', '2.0.0:'): 'astor_archive',
        ('astunparse', '2.2.0:'): 'astunparse_archive',
        ('cython', '2.0.0:'): 'cython',  # Part of Python EC
        ('dill', '2.4.0:'): 'dill_archive',
        ('enum', '2.0.0:2.8.0'): 'enum34_archive',  # Part of Python3
        ('flatbuffers', '2.4.0:'): 'flatbuffers',
        ('functools', '2.0.0:'): 'functools32_archive',  # Part of Python3
        ('gast', '2.0.0:'): 'gast_archive',
        ('google.protobuf', '2.0.0:'): 'com_google_protobuf',
        ('keras_applications', '2.0.0:2.2.0'): 'keras_applications_archive',
        ('opt_einsum', '2.0.0:'): 'opt_einsum_archive',
        ('pasta', '2.0.0:'): 'pasta',
        ('six', '2.0.0:'): 'six_archive',  # Part of Python EC
        ('tblib', '2.4.0:'): 'tblib_archive',
        ('termcolor', '2.0.0:'): 'termcolor_archive',
        ('typing_extensions', '2.4.0:'): 'typing_extensions_archive',
        ('wrapt', '2.0.0:'): 'wrapt',
    }

    dependency_mapping = dict((dep_name, tf_name)
                              for (dep_name, version_range), tf_name in available_system_libs.items()
                              if is_version_ok(version_range))
    python_mapping = dict((pkg_name, tf_name)
                          for (pkg_name, version_range), tf_name in python_system_libs.items()
                          if is_version_ok(version_range))

    if as_valid_libs:
        tf_names = [tf_name for tf_name, version_range in unused_system_libs.items()
                    if is_version_ok(version_range)]
        tf_names.extend(dependency_mapping.values())
        tf_names.extend(python_mapping.values())
        result = '\n'.join(['    "%s",' % name for name in sorted(tf_names)])
    else:
        result = dependency_mapping, python_mapping
    return result


def get_bazel_version():
    """Get the Bazel version as a LooseVersion. Error if not found"""
    version = get_software_version('Bazel')
    if version is None:
        raise EasyBuildError('Failed to determine Bazel version - is it listed as a (build) dependency?')
    return LooseVersion(version)


class EB_TensorFlow(PythonPackage):
    """Support for building/installing TensorFlow."""

    @staticmethod
    def extra_options():
        extra_vars = {
            'path_filter': [[], "List of patterns to be filtered out in paths in $CPATH and $LIBRARY_PATH", CUSTOM],
            'with_jemalloc': [None, "Make TensorFlow use jemalloc (usually enabled by default). " +
                                    "Unsupported starting at TensorFlow 1.12!", CUSTOM],
            'with_mkl_dnn': [None, "Make TensorFlow use Intel MKL-DNN / oneDNN and configure with --config=mkl "
                                   "(enabled by default where supported for TensorFlow versions before 2.4.0)",
                             CUSTOM],
            'with_xla': [None, "Enable XLA JIT compiler for possible runtime optimization of models", CUSTOM],
            'test_script': [None, "Script to test TensorFlow installation with", CUSTOM],
            'test_targets': [[], "List of Bazel targets which should be run during the test step", CUSTOM],
            'test_tag_filters_cpu': ['', "Comma-separated list of tags to filter for during the CPU test step", CUSTOM],
            'test_tag_filters_gpu': ['', "Comma-separated list of tags to filter for during the GPU test step", CUSTOM],
            'testopts_gpu': ['', 'Test options for the GPU test step', CUSTOM],
            'test_max_parallel': [None, "Maximum number of test jobs to run in parallel (GPU tests are limited by " +
                                  "the number of GPUs). Use None (default) to automatically determine a value", CUSTOM],
            'jvm_max_memory': [4096, "Maximum amount of memory in MB used for the JVM running Bazel." +
                               "Use None to not set a specific limit (uses a default value).", CUSTOM],
        }

        return PythonPackage.extra_options(extra_vars)

    def __init__(self, *args, **kwargs):
        """Initialize TensorFlow easyblock."""
        super(EB_TensorFlow, self).__init__(*args, **kwargs)

        with self.cfg.disable_templating():
            self.cfg['exts_defaultclass'] = 'PythonPackage'

            self.cfg['exts_default_options']['download_dep_fail'] = True
            self.cfg['exts_default_options']['use_pip'] = True
            self.cfg['exts_filter'] = EXTS_FILTER_PYTHON_PACKAGES

        self.system_libs_info = None

        self.test_script = None

        # locate test script (if specified)
        if self.cfg['test_script']:
            # try to locate test script via obtain_file (just like sources & patches files)
            self.test_script = self.obtain_file(self.cfg['test_script'])
            if self.test_script and os.path.exists(self.test_script):
                self.log.info("Test script found: %s", self.test_script)
            else:
                raise EasyBuildError("Specified test script %s not found!", self.cfg['test_script'])

    def python_pkg_exists(self, name):
        """Check if the given python package exists/can be imported"""
        cmd = self.python_cmd + " -c 'import %s'" % name
        out, ec = run_cmd(cmd, log_ok=False)
        self.log.debug('Existence check for %s returned %s with output: %s', name, ec, out)
        return ec == 0

    def handle_jemalloc(self):
        """Figure out whether jemalloc support should be enabled or not."""
        if self.cfg['with_jemalloc'] is None:
            if LooseVersion(self.version) > LooseVersion('1.6'):
                # jemalloc bundled with recent versions of TensorFlow does not work on RHEL 6 or derivatives,
                # so disable it automatically if with_jemalloc was left unspecified
                os_name = get_os_name().replace(' ', '')
                rh_based_os = any(os_name.startswith(x) for x in ['centos', 'redhat', 'rhel', 'sl'])
                if rh_based_os and get_os_version().startswith('6.'):
                    self.log.info("Disabling jemalloc since bundled jemalloc does not work on RHEL 6 and derivatives")
                    self.cfg['with_jemalloc'] = False

            # if the above doesn't disable jemalloc support, then enable it by default
            if self.cfg['with_jemalloc'] is None:
                self.log.info("Enabling jemalloc support by default, since it was left unspecified")
                self.cfg['with_jemalloc'] = True

        else:
            # if with_jemalloc was specified, stick to that
            self.log.info("with_jemalloc was specified as %s, so sticking to it", self.cfg['with_jemalloc'])

    def write_wrapper(self, wrapper_dir, compiler, i_mpi_root):
        """Helper function to write a compiler wrapper."""
        wrapper_txt = INTEL_COMPILER_WRAPPER % {
            'compiler_path': which(compiler),
            'intel_mpi_root': i_mpi_root,
            'cpath': os.getenv('CPATH'),
            'intel_license_file': os.getenv('INTEL_LICENSE_FILE', os.getenv('LM_LICENSE_FILE')),
            'wrapper_dir': wrapper_dir,
        }
        wrapper = os.path.join(wrapper_dir, compiler)
        write_file(wrapper, wrapper_txt)
        if self.dry_run:
            self.dry_run_msg("Wrapper for '%s' was put in place: %s", compiler, wrapper)
        else:
            adjust_permissions(wrapper, stat.S_IXUSR)
            self.log.info("Using wrapper script for '%s': %s", compiler, which(compiler))

    def verify_system_libs_info(self):
        """Verifies that the stored info about $TF_SYSTEM_LIBS is complete"""
        available_libs_src = set(get_system_libs_from_tf(self.start_dir))
        available_libs_eb = set(split_tf_libs_txt(get_system_libs_for_version(self.version, as_valid_libs=True)))
        # If available_libs_eb is empty it is not an error e.g. it is not worth trying to make all old ECs work
        # So we just log it so it can be verified manually if required
        if not available_libs_eb:
            self.log.warning('TensorFlow EasyBlock does not have any information for $TF_SYSTEM_LIBS stored. ' +
                             'This means most dependencies will be downloaded at build time by TensorFlow.\n' +
                             'Available $TF_SYSTEM_LIBS according to the TensorFlow sources: %s',
                             sorted(available_libs_src))
            return
        # Those 2 sets should be equal. We determine the differences here to report better errors
        missing_libs = available_libs_src - available_libs_eb
        unknown_libs = available_libs_eb - available_libs_src
        if missing_libs or unknown_libs:
            if not available_libs_src:
                msg = 'Failed to determine available $TF_SYSTEM_LIBS from the source'
            else:
                msg = 'Values for $TF_SYSTEM_LIBS in the TensorFlow EasyBlock are incomplete.\n'
                if missing_libs:
                    # Libs available according to TF sources but not listed in this EasyBlock
                    msg += 'Missing entries for $TF_SYSTEM_LIBS: %s\n' % missing_libs
                if unknown_libs:
                    # Libs listed in this EasyBlock but not present in the TF sources -> Removed?
                    msg += 'Unrecognized entries for $TF_SYSTEM_LIBS: %s\n' % unknown_libs
                msg += 'The EasyBlock needs to be updated to fully work with TensorFlow version %s' % self.version
            if build_option('strict') == run.ERROR:
                raise EasyBuildError(msg)
            else:
                print_warning(msg)

    def get_system_libs(self):
        """
        Get list of dependencies for $TF_SYSTEM_LIBS

        Returns a tuple of lists: $TF_SYSTEM_LIBS names, include paths, library paths
        """
        dependency_mapping, python_mapping = get_system_libs_for_version(self.version)
        # Some TF dependencies require both a (usually C++) dependency and a Python package
        deps_with_python_pkg = set(tf_name for tf_name in dependency_mapping.values()
                                   if tf_name in python_mapping.values())

        system_libs = []
        cpaths = []
        libpaths = []
        ignored_system_deps = []

        # Check direct dependencies
        dep_names = set(dep['name'] for dep in self.cfg.dependencies())
        for dep_name, tf_name in sorted(dependency_mapping.items(), key=lambda i: i[0].lower()):
            if dep_name in dep_names:
                if tf_name in deps_with_python_pkg:
                    pkg_name = next(cur_pkg_name for cur_pkg_name, cur_tf_name in python_mapping.items()
                                    if cur_tf_name == tf_name)
                    # Simply ignore. Error reporting is done in the other loop
                    if not self.python_pkg_exists(pkg_name):
                        continue
                system_libs.append(tf_name)
                # When using cURL (which uses the system OpenSSL), we also need to use "boringssl"
                # which essentially resolves to using OpenSSL as the API and library names are compatible
                if dep_name == 'cURL':
                    system_libs.append('boringssl')
                sw_root = get_software_root(dep_name)
                # Dependency might be filtered via --filter-deps. In that case assume globally installed version
                if not sw_root:
                    continue
                incpath = os.path.join(sw_root, 'include')
                if os.path.exists(incpath):
                    cpaths.append(incpath)
                    if dep_name == 'JsonCpp' and LooseVersion(self.version) < LooseVersion('2.3'):
                        # Need to add the install prefix or patch the sources:
                        # https://github.com/tensorflow/tensorflow/issues/42303
                        cpaths.append(sw_root)
                    if dep_name == 'protobuf':
                        if LooseVersion(self.version) < LooseVersion('2.4'):
                            # Need to set INCLUDEDIR as TF wants to symlink files from there:
                            # https://github.com/tensorflow/tensorflow/issues/37835
                            env.setvar('INCLUDEDIR', incpath)
                        else:
                            env.setvar('PROTOBUF_INCLUDE_PATH', incpath)
                libpath = get_software_libdir(dep_name)
                if libpath:
                    libpaths.append(os.path.join(sw_root, libpath))
            else:
                ignored_system_deps.append('%s (Dependency %s)' % (tf_name, dep_name))

        for pkg_name, tf_name in sorted(python_mapping.items(), key=lambda i: i[0].lower()):
            if self.python_pkg_exists(pkg_name):
                # If it is in deps_with_python_pkg we already added it
                if tf_name not in deps_with_python_pkg:
                    system_libs.append(tf_name)
            else:
                ignored_system_deps.append('%s (Python package %s)' % (tf_name, pkg_name))

        # If we use OpenSSL (potentially as a wrapper) somewhere in the chain we must tell TF to use it too
        openssl_root = get_software_root('OpenSSL')
        if openssl_root:
            if 'boringssl' not in system_libs:
                system_libs.append('boringssl')
            incpath = os.path.join(openssl_root, 'include')
            if os.path.exists(incpath):
                cpaths.append(incpath)
            libpath = get_software_libdir(dep_name)
            if libpath:
                libpaths.append(os.path.join(openssl_root, libpath))

        if ignored_system_deps:
            print_warning('%d TensorFlow dependencies have not been resolved by EasyBuild. Check the log for details.',
                          len(ignored_system_deps))
            self.log.warning('For the following $TF_SYSTEM_LIBS dependencies TensorFlow will download a copy ' +
                             'because an EB dependency was not found: \n%s\n' +
                             'EC Dependencies: %s\n' +
                             'Installed Python packages: %s\n',
                             ', '.join(ignored_system_deps),
                             ', '.join(dep_names),
                             ', '.join(self.get_installed_python_packages()))
        else:
            self.log.info("All known TensorFlow $TF_SYSTEM_LIBS dependencies resolved via EasyBuild!")

        return system_libs, cpaths, libpaths

    def setup_build_dirs(self):
        """Setup temporary build directories"""
        # This is either the builddir (for standalone builds) or the extension sub folder when TF is an extension
        # Either way this folder only contains the folder with the sources and hence we can use fixed names
        # for the subfolders
        parent_dir = os.path.dirname(self.start_dir)
        # Path where Bazel will store its output, build artefacts etc.
        self.output_user_root_dir = os.path.join(parent_dir, 'bazel-root')
        # Folder where wrapper binaries can be placed, where required. TODO: Replace by --action_env cmds
        self.wrapper_dir = os.path.join(parent_dir, 'wrapper_bin')

    def configure_step(self):
        """Custom configuration procedure for TensorFlow."""

        # Bazel seems to not be able to handle a large amount of parallel jobs, e.g. 176 on some Power machines,
        # and will hang forever building the TensorFlow package.
        # So limit to something high but still reasonable while allowing ECs to overwrite it
        if self.cfg['maxparallel'] is None:
            # Seemingly Bazel around 3.x got better, so double the max there
            bazel_max = 64 if get_bazel_version() < '3.0.0' else 128
            self.cfg['parallel'] = min(self.cfg['parallel'], bazel_max)

        binutils_root = get_software_root('binutils')
        if not binutils_root:
            raise EasyBuildError("Failed to determine installation prefix for binutils")
        self.binutils_bin_path = os.path.join(binutils_root, 'bin')

        # filter out paths from CPATH and LIBRARY_PATH. This is needed since bazel will pull some dependencies that
        # might conflict with dependencies on the system and/or installed with EB. For example: protobuf
        path_filter = self.cfg['path_filter']
        if path_filter:
            self.log.info("Filtering $CPATH and $LIBRARY_PATH with path filter %s", path_filter)
            for var in ['CPATH', 'LIBRARY_PATH']:
                path = os.getenv(var).split(os.pathsep)
                self.log.info("$%s old value was %s" % (var, path))
                filtered_path = os.pathsep.join([p for fil in path_filter for p in path if fil not in p])
                env.setvar(var, filtered_path)

        self.setup_build_dirs()

        use_wrapper = False
        if self.toolchain.comp_family() == toolchain.INTELCOMP:
            # put wrappers for Intel C/C++ compilers in place (required to make sure license server is found)
            # cfr. https://github.com/bazelbuild/bazel/issues/663
            for compiler in ('icc', 'icpc'):
                self.write_wrapper(self.wrapper_dir, compiler, 'NOT-USED-WITH-ICC')
            use_wrapper = True

        use_mpi = self.toolchain.options.get('usempi', False)
        mpi_home = ''
        if use_mpi:
            impi_root = get_software_root('impi')
            if impi_root:
                # put wrappers for Intel MPI compiler wrappers in place
                # (required to make sure license server and I_MPI_ROOT are found)
                for compiler in (os.getenv('MPICC'), os.getenv('MPICXX')):
                    self.write_wrapper(self.wrapper_dir, compiler, os.getenv('I_MPI_ROOT'))
                use_wrapper = True
                # set correct value for MPI_HOME
                mpi_home = os.path.join(impi_root, 'intel64')
            else:
                self.log.debug("MPI module name: %s", self.toolchain.MPI_MODULE_NAME[0])
                mpi_home = get_software_root(self.toolchain.MPI_MODULE_NAME[0])

            self.log.debug("Derived value for MPI_HOME: %s", mpi_home)

        if use_wrapper:
            env.setvar('PATH', os.pathsep.join([self.wrapper_dir, os.getenv('PATH')]))

        self.prepare_python()

        self.verify_system_libs_info()
        self.system_libs_info = self.get_system_libs()

        # Options passed to the target (build/test), e.g. --config arguments
        self.target_opts = []

        cuda_root = get_software_root('CUDA')
        cudnn_root = get_software_root('cuDNN')
        opencl_root = get_software_root('OpenCL')
        tensorrt_root = get_software_root('TensorRT')
        nccl_root = get_software_root('NCCL')

        self._with_cuda = bool(cuda_root)

        config_env_vars = {
            'CC_OPT_FLAGS': os.getenv('CXXFLAGS'),
            'MPI_HOME': mpi_home,
            'PYTHON_BIN_PATH': self.python_cmd,
            'PYTHON_LIB_PATH': os.path.join(self.installdir, self.pylibdir),
            'TF_CUDA_CLANG': '0',
            'TF_DOWNLOAD_CLANG': '0',  # Still experimental in TF 2.1.0
            'TF_ENABLE_XLA': ('0', '1')[bool(self.cfg['with_xla'])],  # XLA JIT support
            'TF_NEED_CUDA': ('0', '1')[self._with_cuda],
            'TF_NEED_OPENCL': ('0', '1')[bool(opencl_root)],
            'TF_NEED_ROCM': '0',
            'TF_NEED_TENSORRT': '0',
            'TF_SET_ANDROID_WORKSPACE': '0',
            'TF_SYSTEM_LIBS': ','.join(self.system_libs_info[0]),
        }
        if LooseVersion(self.version) < LooseVersion('1.10'):
            config_env_vars['TF_NEED_S3'] = '0'  # Renamed to TF_NEED_AWS in 1.9.0-rc2 and 1.10, not 1.9.0
        # Options removed in 1.12.0
        if LooseVersion(self.version) < LooseVersion('1.12'):
            self.handle_jemalloc()
            config_env_vars.update({
                'TF_NEED_AWS': '0',  # Amazon AWS Platform
                'TF_NEED_GCP': '0',  # Google Cloud Platform
                'TF_NEED_GDR': '0',
                'TF_NEED_HDFS': '0',  # Hadoop File System
                'TF_NEED_JEMALLOC': ('0', '1')[self.cfg['with_jemalloc']],
                'TF_NEED_KAFKA': '0',  # Amazon Kafka Platform
                'TF_NEED_VERBS': '0',
            })
        elif self.cfg['with_jemalloc'] is True:
            print_warning('Jemalloc is not supported in TensorFlow %s, the EC option with_jemalloc has no effect',
                          self.version)
        # Disable support of some features via config switch introduced in 1.12.1
        if LooseVersion(self.version) >= LooseVersion('1.12.1'):
            self.target_opts += ['--config=noaws', '--config=nogcp', '--config=nohdfs']
            # Removed in 2.1
            if LooseVersion(self.version) < LooseVersion('2.1'):
                self.target_opts.append('--config=nokafka')
        # MPI support removed in 2.1
        if LooseVersion(self.version) < LooseVersion('2.1'):
            config_env_vars['TF_NEED_MPI'] = ('0', '1')[bool(use_mpi)]
        # SYCL support removed in 2.4
        if LooseVersion(self.version) < LooseVersion('2.4'):
            config_env_vars['TF_NEED_OPENCL_SYCL'] = '0'

        if self._with_cuda:
            cuda_version = get_software_version('CUDA')
            cuda_maj_min_ver = '.'.join(cuda_version.split('.')[:2])

            # $GCC_HOST_COMPILER_PATH should be set to path of the actual compiler (not the MPI compiler wrapper)
            if use_mpi:
                compiler_path = which(os.getenv('CC_SEQ'))
            else:
                compiler_path = which(os.getenv('CC'))

            # list of CUDA compute capabilities to use can be specifed in two ways (where (2) overrules (1)):
            # (1) in the easyconfig file, via the custom cuda_compute_capabilities;
            # (2) in the EasyBuild configuration, via --cuda-compute-capabilities configuration option;
            ec_cuda_cc = self.cfg['cuda_compute_capabilities']
            cfg_cuda_cc = build_option('cuda_compute_capabilities')
            cuda_cc = cfg_cuda_cc or ec_cuda_cc or []

            if cfg_cuda_cc and ec_cuda_cc:
                warning_msg = "cuda_compute_capabilities specified in easyconfig (%s) are overruled by " % ec_cuda_cc
                warning_msg += "--cuda-compute-capabilities configuration option (%s)" % cfg_cuda_cc
                print_warning(warning_msg)
            elif not cuda_cc:
                warning_msg = "No CUDA compute capabilities specified, so using TensorFlow default "
                warning_msg += "(which may not be optimal for your system).\nYou should use "
                warning_msg += "the --cuda-compute-capabilities configuration option or the cuda_compute_capabilities "
                warning_msg += "easyconfig parameter to specify a list of CUDA compute capabilities to compile with."
                print_warning(warning_msg)

            # TensorFlow 1.12.1 requires compute capability >= 3.5
            # see https://github.com/tensorflow/tensorflow/pull/25767
            if LooseVersion(self.version) >= LooseVersion('1.12.1'):
                faulty_comp_caps = [x for x in cuda_cc if LooseVersion(x) < LooseVersion('3.5')]
                if faulty_comp_caps:
                    error_msg = "TensorFlow >= 1.12.1 requires CUDA compute capabilities >= 3.5, "
                    error_msg += "found one or more older ones: %s"
                    raise EasyBuildError(error_msg, ', '.join(faulty_comp_caps))

            if cuda_cc:
                self.log.info("Compiling with specified list of CUDA compute capabilities: %s", ', '.join(cuda_cc))

            config_env_vars.update({
                'CUDA_TOOLKIT_PATH': cuda_root,
                'GCC_HOST_COMPILER_PATH': compiler_path,
                # This is the binutils bin folder: https://github.com/tensorflow/tensorflow/issues/39263
                'GCC_HOST_COMPILER_PREFIX': self.binutils_bin_path,
                'TF_CUDA_COMPUTE_CAPABILITIES': ','.join(cuda_cc),
                'TF_CUDA_VERSION': cuda_maj_min_ver,
            })

            # for recent TensorFlow versions, $TF_CUDA_PATHS and $TF_CUBLAS_VERSION must also be set
            if LooseVersion(self.version) >= LooseVersion('1.14'):

                # figure out correct major/minor version for CUBLAS from cublas_api.h
                cublas_api_header_glob_pattern = os.path.join(cuda_root, 'targets', '*', 'include', 'cublas_api.h')
                matches = glob.glob(cublas_api_header_glob_pattern)
                if len(matches) == 1:
                    cublas_api_header_path = matches[0]
                    cublas_api_header_txt = read_file(cublas_api_header_path)
                else:
                    raise EasyBuildError("Failed to isolate path to cublas_api.h: %s", matches)

                cublas_ver_parts = []
                for key in ['CUBLAS_VER_MAJOR', 'CUBLAS_VER_MINOR', 'CUBLAS_VER_PATCH']:
                    regex = re.compile("^#define %s ([0-9]+)" % key, re.M)
                    res = regex.search(cublas_api_header_txt)
                    if res:
                        cublas_ver_parts.append(res.group(1))
                    else:
                        raise EasyBuildError("Failed to find pattern '%s' in %s", regex.pattern, cublas_api_header_path)

                config_env_vars.update({
                    'TF_CUDA_PATHS': cuda_root,
                    'TF_CUBLAS_VERSION': '.'.join(cublas_ver_parts),
                })

            if cudnn_root:
                cudnn_version = get_software_version('cuDNN')
                cudnn_maj_min_patch_ver = '.'.join(cudnn_version.split('.')[:3])

                config_env_vars.update({
                    'CUDNN_INSTALL_PATH': cudnn_root,
                    'TF_CUDNN_VERSION': cudnn_maj_min_patch_ver,
                })
            else:
                raise EasyBuildError("TensorFlow has a strict dependency on cuDNN if CUDA is enabled")
            if nccl_root:
                nccl_version = get_software_version('NCCL')
                # Ignore the PKG_REVISION identifier if it exists (i.e., report 2.4.6 for 2.4.6-1 or 2.4.6-2)
                nccl_version = nccl_version.split('-')[0]
                config_env_vars.update({
                    'NCCL_INSTALL_PATH': nccl_root,
                })
            else:
                nccl_version = '1.3'  # Use simple downloadable version
            config_env_vars.update({
                'TF_NCCL_VERSION': nccl_version,
            })
            if tensorrt_root:
                tensorrt_version = get_software_version('TensorRT')
                config_env_vars.update({
                    'TF_NEED_TENSORRT': '1',
                    'TENSORRT_INSTALL_PATH': tensorrt_root,
                    'TF_TENSORRT_VERSION': tensorrt_version,
                })

        configure_py_contents = read_file('configure.py')
        for key, val in sorted(config_env_vars.items()):
            if key.startswith('TF_') and key not in configure_py_contents:
                self.log.warn('Did not find %s option in configure.py. Setting might not have any effect', key)
            env.setvar(key, val)

        # configure.py (called by configure script) already calls bazel to determine the bazel version
        # Since 2.3.0 `bazel --version` is used which doesn't extract bazel, prior it did
        # Hence make sure it doesn't extract into $HOME/.cache/bazel
        if LooseVersion(self.version) < LooseVersion('2.3.0'):
            regex_subs = [(r"('bazel', '--batch')",
                           r"\1, '--output_user_root=%s'" % self.output_user_root_dir)]
            apply_regex_substitutions('configure.py', regex_subs)

        cmd = self.cfg['preconfigopts'] + './configure ' + self.cfg['configopts']
        run_cmd(cmd, log_all=True, simple=True)

    def patch_crosstool_files(self):
        """Patches the CROSSTOOL files to include EasyBuild provided compiler paths"""
        inc_paths, lib_paths = [], []

        gcc_root = get_software_root('GCCcore') or get_software_root('GCC')
        if gcc_root:
            gcc_lib64 = os.path.join(gcc_root, 'lib64')
            lib_paths.append(gcc_lib64)

            gcc_ver = get_software_version('GCCcore') or get_software_version('GCC')

            # figure out location of GCC include files
            # make sure we don't pick up the nvptx-none directory by looking for a specific include file
            res = glob.glob(os.path.join(gcc_root, 'lib', 'gcc', '*', gcc_ver, 'include', 'immintrin.h'))
            if res and len(res) == 1:
                gcc_lib_inc = os.path.dirname(res[0])
                inc_paths.append(gcc_lib_inc)
            else:
                raise EasyBuildError("Failed to pinpoint location of GCC include files: %s", res)

            # make sure include-fixed directory is where we expect it to be
            gcc_lib_inc_fixed = os.path.join(os.path.dirname(gcc_lib_inc), 'include-fixed')
            if os.path.exists(gcc_lib_inc_fixed):
                inc_paths.append(gcc_lib_inc_fixed)
            else:
                self.log.info("Derived directory %s does not exist, so discarding it", gcc_lib_inc_fixed)

            # also check on location of include/c++/<gcc version> directory
            gcc_cplusplus_inc = os.path.join(gcc_root, 'include', 'c++', gcc_ver)
            if os.path.exists(gcc_cplusplus_inc):
                inc_paths.append(gcc_cplusplus_inc)
            else:
                raise EasyBuildError("Derived directory %s does not exist", gcc_cplusplus_inc)
        else:
            raise EasyBuildError("Failed to determine installation prefix for GCC")

        cuda_root = get_software_root('CUDA')
        if cuda_root:
            inc_paths.append(os.path.join(cuda_root, 'include'))
            lib_paths.append(os.path.join(cuda_root, 'lib64'))

        # fix hardcoded locations of compilers & tools
        cxx_inc_dirs = ['cxx_builtin_include_directory: "%s"' % resolve_path(p) for p in inc_paths]
        cxx_inc_dirs += ['cxx_builtin_include_directory: "%s"' % p for p in inc_paths]
        regex_subs = [
            (r'-B/usr/bin/', '-B%s %s' % (self.binutils_bin_path, ' '.join('-L%s/' % p for p in lib_paths))),
            (r'(cxx_builtin_include_directory:).*', ''),
            (r'^toolchain {', 'toolchain {\n' + '\n'.join(cxx_inc_dirs)),
        ]
        for tool in ['ar', 'cpp', 'dwp', 'gcc', 'gcov', 'ld', 'nm', 'objcopy', 'objdump', 'strip']:
            path = which(tool)
            if path:
                regex_subs.append((os.path.join('/usr', 'bin', tool), path))
            else:
                raise EasyBuildError("Failed to determine path to '%s'", tool)

        # -fPIE/-pie and -fPIC are not compatible, so patch out hardcoded occurences of -fPIE/-pie if -fPIC is used
        if self.toolchain.options.get('pic', None):
            regex_subs.extend([('-fPIE', '-fPIC'), ('"-pie"', '"-fPIC"')])

        # patch all CROSSTOOL* scripts to fix hardcoding of locations of binutils/GCC binaries
        for path, dirnames, filenames in os.walk(os.getcwd()):
            for filename in filenames:
                if filename.startswith('CROSSTOOL'):
                    full_path = os.path.join(path, filename)
                    self.log.info("Patching %s", full_path)
                    apply_regex_substitutions(full_path, regex_subs)

    def build_step(self):
        """Custom build procedure for TensorFlow."""

        bazel_version = get_bazel_version()

        # pre-create target installation directory
        mkdir(os.path.join(self.installdir, self.pylibdir), parents=True)

        # This seems to be no longer required since at least 2.0, likely also for older versions
        if LooseVersion(self.version) < LooseVersion('2.0'):
            self.patch_crosstool_files()

        # Options passed to the bazel command
        self.bazel_opts = [
            '--output_user_root=%s' % self.output_user_root_dir,
        ]
        # Increase time to wait for bazel to start, available since 4.0+
        if bazel_version >= '4.0.0':
            self.bazel_opts.append('--local_startup_timeout_secs=300')  # 5min

        # Environment variables and values needed for Bazel actions.
        action_env = {}
        # A value of None is interpreted as using the invoking environments value
        INHERIT = None  # For better readability

        jvm_max_memory = self.cfg['jvm_max_memory']
        if jvm_max_memory:
            jvm_startup_memory = min(512, int(jvm_max_memory))
            self.bazel_opts.extend([
                '--host_jvm_args=-Xms%sm' % jvm_startup_memory,
                '--host_jvm_args=-Xmx%sm' % jvm_max_memory
            ])

        if self.toolchain.options.get('debug', None):
            self.target_opts.append('--strip=never')
            self.target_opts.append('--compilation_mode=dbg')
            self.target_opts.append('--copt="-Og"')
        else:
            # build with optimization enabled
            # cfr. https://docs.bazel.build/versions/master/user-manual.html#flag--compilation_mode
            self.target_opts.append('--compilation_mode=opt')

            # select 'opt' config section (this is *not* the same as --compilation_mode=opt!)
            # https://docs.bazel.build/versions/master/user-manual.html#flag--config
            self.target_opts.append('--config=opt')

        # make Bazel print full command line + make it verbose on failures
        # https://docs.bazel.build/versions/master/user-manual.html#flag--subcommands
        # https://docs.bazel.build/versions/master/user-manual.html#flag--verbose_failures
        self.target_opts.extend(['--subcommands', '--verbose_failures'])

        self.target_opts.append('--jobs=%s' % self.cfg['parallel'])

        if self.toolchain.options.get('pic', None):
            self.target_opts.append('--copt="-fPIC"')

        # include install location of Python packages in $PYTHONPATH,
        # and specify that value of $PYTHONPATH should be passed down into Bazel build environment;
        # this is required to make sure that Python packages included as extensions are found at build time;
        # see also https://github.com/tensorflow/tensorflow/issues/22395
        pythonpath = os.getenv('PYTHONPATH', '')
        env.setvar('PYTHONPATH', os.pathsep.join([os.path.join(self.installdir, self.pylibdir), pythonpath]))

        # Make TF find our modules. LD_LIBRARY_PATH gets automatically added by configure.py
        cpaths, libpaths = self.system_libs_info[1:]
        if cpaths:
            action_env['CPATH'] = ':'.join(cpaths)
        if libpaths:
            action_env['LIBRARY_PATH'] = ':'.join(libpaths)
        action_env['PYTHONPATH'] = INHERIT
        # Also export $EBPYTHONPREFIXES to handle the multi-deps python setup
        # See https://github.com/easybuilders/easybuild-easyblocks/pull/1664
        if 'EBPYTHONPREFIXES' in os.environ:
            action_env['EBPYTHONPREFIXES'] = INHERIT

        # Ignore user environment for Python
        action_env['PYTHONNOUSERSITE'] = '1'

        # TF 2 (final) sets this in configure
        if LooseVersion(self.version) < LooseVersion('2.0'):
            if self._with_cuda:
                self.target_opts.append('--config=cuda')

        # note: using --config=mkl results in a significantly different build, with a different
        # threading model (which may lead to thread oversubscription and significant performance loss,
        # see https://github.com/easybuilders/easybuild-easyblocks/issues/2577) and different
        # runtime behavior w.r.t. GPU vs CPU execution of functions like tf.matmul
        # (see https://github.com/easybuilders/easybuild-easyconfigs/issues/14120),
        # so make sure you really know you want to use this!

        # auto-enable use of MKL-DNN/oneDNN and --config=mkl when possible if with_mkl_dnn is left unspecified;
        # only do this for TensorFlow versions older than 2.4.0, since more recent versions
        # oneDNN is used automatically for x86_64 systems (and mkl-dnn is no longer a dependency);
        if self.cfg['with_mkl_dnn'] is None and LooseVersion(self.version) < LooseVersion('2.4.0'):
            cpu_arch = get_cpu_architecture()
            if cpu_arch == X86_64:
                # Supported on x86 since forever
                self.cfg['with_mkl_dnn'] = True
                self.log.info("Auto-enabled use of MKL-DNN on %s CPU architecture", cpu_arch)
            else:
                self.log.info("Not enabling use of MKL-DNN on %s CPU architecture", cpu_arch)

        # if mkl-dnn is listed as a dependency it is used
        mkl_root = get_software_root('mkl-dnn')
        if mkl_root:
            self.target_opts.append('--config=mkl')
            env.setvar('TF_MKL_ROOT', mkl_root)
        elif self.cfg['with_mkl_dnn']:
            # this makes TensorFlow use mkl-dnn (cfr. https://github.com/01org/mkl-dnn),
            # and download it if needed
            self.target_opts.append('--config=mkl')

        # Use the same configuration (i.e. environment) for compiling and using host tools
        # This means that our action_envs are (almost) always passed
        # Fully removed in Bazel 6.0 and limited effect after at least 3.7 (see --host_action_env)
        if bazel_version < '6.0.0':
            self.target_opts.append('--distinct_host_configuration=false')

        for key, value in sorted(action_env.items()):
            option = key
            if value is not None:
                option += "='%s'" % value

            self.target_opts.append('--action_env=' + option)
            if bazel_version >= '3.7.0':
                # Since Bazel 3.7 action_env only applies to the "target" environment, not the "host" environment
                # As we are not cross-compiling we need both be the same -> Duplicate the setting to host_action_env
                # See https://github.com/bazelbuild/bazel/commit/a463d9095386b22c121d20957222dbb44caef7d4
                self.target_opts.append('--host_action_env=' + option)

        # Compose final command
        cmd = (
            [self.cfg['prebuildopts']]
            + ['bazel']
            + self.bazel_opts
            + ['build']
            + self.target_opts
            + [self.cfg['buildopts']]
            # specify target of the build command as last argument
            + ['//tensorflow/tools/pip_package:build_pip_package']
        )

        run_cmd(' '.join(cmd), log_all=True, simple=True, log_ok=True)

        # run generated 'build_pip_package' script to build the .whl
        cmd = "bazel-bin/tensorflow/tools/pip_package/build_pip_package %s" % self.builddir
        run_cmd(cmd, log_all=True, simple=True, log_ok=True)

    def test_step(self):
        """Run TensorFlow unit tests"""
        # IMPORTANT: This code allows experiments with running TF tests but may change
        test_targets = self.cfg['test_targets']
        if not test_targets:
            self.log.info('No targets selected for tests. Set e.g. test_targets = ["//tensorflow/python/..."] '
                          'to run TensorFlow tests.')
            return
        # Allow a string as the test_targets (useful for C&P testing from TF sources)
        if not isinstance(test_targets, list):
            test_targets = test_targets.split(' ')

        test_opts = self.target_opts
        test_opts.append('--test_output=errors')  # (Additionally) show logs from failed tests
        test_opts.append('--build_tests_only')  # Don't build tests which won't be executed

        # determine number of cores/GPUs to use for tests
        max_num_test_jobs = int(self.cfg['test_max_parallel'] or self.cfg['parallel'])
        if self._with_cuda:
            if not which('nvidia-smi', on_error=IGNORE):
                print_warning('Could not find nvidia-smi. Assuming a system without GPUs and skipping GPU tests!')
                num_gpus_to_use = 0
            elif os.environ.get('CUDA_VISIBLE_DEVICES') == '-1':
                print_warning('GPUs explicitely disabled via CUDA_VISIBLE_DEVICES. Skipping GPU tests!')
                num_gpus_to_use = 0
            else:
                # determine number of available GPUs via nvidia-smi command, fall back to just 1 GPU
                # Note: Disable logging to also disable the error handling in run_cmd and do it explicitly below
                (out, ec) = run_cmd("nvidia-smi --list-gpus", log_ok=False, log_all=False, regexp=False)
                try:
                    if ec != 0:
                        raise RuntimeError("nvidia-smi returned exit code %s with output:\n%s" % (ec, out))
                    else:
                        self.log.info('nvidia-smi succeeded with output:\n%s' % out)
                        gpu_ct = sum(line.startswith('GPU ') for line in out.strip().split('\n'))
                except (RuntimeError, ValueError) as err:
                    self.log.warning("Failed to get the number of GPUs on this system: %s", err)
                    gpu_ct = 0

                if gpu_ct == 0:
                    print_warning('No GPUs found. Skipping GPU tests!')

                num_gpus_to_use = min(max_num_test_jobs, gpu_ct)

            # Can (likely) only run 1 test per GPU but don't need to limit CPU tests
            num_test_jobs = {
                CPU_DEVICE: max_num_test_jobs,
                GPU_DEVICE: num_gpus_to_use,
            }
        else:
            num_test_jobs = {
                CPU_DEVICE: max_num_test_jobs,
                GPU_DEVICE: 0,
            }

        cfg_testopts = {
            CPU_DEVICE: self.cfg['testopts'],
            GPU_DEVICE: self.cfg['testopts_gpu'],
        }

        devices = [CPU_DEVICE]
        # Skip GPU tests if not build with CUDA or no test jobs set (e.g. due to no GPUs available)
        if self._with_cuda and num_test_jobs[GPU_DEVICE]:
            devices.append(GPU_DEVICE)

        for device in devices:
            # Determine tests to run
            test_tag_filters_name = 'test_tag_filters_' + device
            test_tag_filters = self.cfg[test_tag_filters_name]
            if not test_tag_filters:
                self.log.info('Skipping %s test because %s is not set', device, test_tag_filters_name)
                continue
            else:
                self.log.info('Starting %s test', device)

            current_test_opts = test_opts[:]
            current_test_opts.append('--local_test_jobs=%s' % num_test_jobs[device])

            # Add both build and test tag filters as done by the TF CI scripts
            current_test_opts.extend("--%s_tag_filters='%s'" % (step, test_tag_filters) for step in ('test', 'build'))

            # Disable all GPUs for the CPU tests, by setting $CUDA_VISIBLE_DEVICES to -1,
            # otherwise TensorFlow will still use GPUs and fail.
            # Only tests explicitely marked with the 'gpu' tag can run with GPUs visible;
            # see https://github.com/tensorflow/tensorflow/issues/45664
            if device == CPU_DEVICE:
                current_test_opts.append("--test_env=CUDA_VISIBLE_DEVICES='-1'")
            else:
                # Propagate those environment variables to the GPU tests if they are set
                important_cuda_env_vars = (
                    'CUDA_CACHE_DISABLE',
                    'CUDA_CACHE_MAXSIZE',
                    'CUDA_CACHE_PATH',
                    'CUDA_FORCE_PTX_JIT',
                    'CUDA_DISABLE_PTX_JIT'
                )
                current_test_opts.extend(
                    '--test_env=' + var_name
                    for var_name in important_cuda_env_vars
                    if var_name in os.environ
                )

                # These are used by the `parallel_gpu_execute` helper script from TF
                current_test_opts.append('--test_env=TF_GPU_COUNT=%s' % num_test_jobs[GPU_DEVICE])
                current_test_opts.append('--test_env=TF_TESTS_PER_GPU=1')

            # Append user specified options last
            current_test_opts.append(cfg_testopts[device])

            # Compose final command
            cmd = ' '.join(
                [self.cfg['pretestopts']]
                + ['bazel']
                + self.bazel_opts
                + ['test']
                + current_test_opts
                + ['--']
                # specify targets to test as last argument
                + test_targets
            )

            stdouterr, ec = run_cmd(cmd, log_ok=False, simple=False)
            if ec:
                fail_msg = 'Tests on %s (cmd: %s) failed with exit code %s and output:\n%s' % (
                    device, cmd, ec, stdouterr)
                self.log.warning(fail_msg)
                # Try to enhance error message
                failed_tests = []
                failed_test_logs = dict()
                # Bazel outputs failed tests like "//tensorflow/c:kernels_test   FAILED in[...]"
                for match in re.finditer(r'^(//[a-zA-Z_/:]+)\s+FAILED', stdouterr, re.MULTILINE):
                    test_name = match.group(1)
                    failed_tests.append(test_name)
                    # Logs are in a folder named after the test, e.g. tensorflow/c/kernels_test
                    test_folder = test_name[2:].replace(':', '/')
                    # Example file names:
                    # <prefix>/k8-opt/testlogs/tensorflow/c/kernels_test/test.log
                    # <prefix>/k8-opt/testlogs/tensorflow/c/kernels_test/shard_1_of_4/test_attempts/attempt_1.log
                    test_log_re = re.compile(r'.*\n(.*\n)?\s*(/.*/testlogs/%s/(/[^/]*)?test.log)' % test_folder)
                    log_match = test_log_re.match(stdouterr, match.end())
                    if log_match:
                        failed_test_logs[test_name] = log_match.group(2)
                # When TF logs are found enhance the below error by additionally logging the details about failed tests
                for test_name, log_path in failed_test_logs.items():
                    if os.path.exists(log_path):
                        self.log.warning('Test %s failed with output\n%s', test_name,
                                         read_file(log_path, log_error=False))
                if failed_tests:
                    failed_tests = sorted(set(failed_tests))  # Make unique to not count retries
                    fail_msg = 'At least %s %s tests failed:\n%s' % (
                        len(failed_tests), device, ', '.join(failed_tests))
                self.report_test_failure(fail_msg)
            else:
                self.log.info('Tests on %s succeeded with output:\n%s', device, stdouterr)

    def install_step(self):
        """Custom install procedure for TensorFlow."""
        # find .whl file that was built, and install it using 'pip install'
        if ("-rc" in self.version):
            whl_version = self.version.replace("-rc", "rc")
        else:
            whl_version = self.version

        whl_paths = glob.glob(os.path.join(self.builddir, 'tensorflow-%s-*.whl' % whl_version))
        if not whl_paths:
            whl_paths = glob.glob(os.path.join(self.builddir, 'tensorflow-*.whl'))
        if len(whl_paths) == 1:
            # --ignore-installed is required to ensure *this* wheel is installed
            cmd = "pip install --ignore-installed --prefix=%s %s" % (self.installdir, whl_paths[0])

            # if extensions are listed, assume they will provide all required dependencies,
            # so use --no-deps to prevent pip from downloading & installing them
            if self.cfg['exts_list']:
                cmd += ' --no-deps'

            run_cmd(cmd, log_all=True, simple=True, log_ok=True)
        else:
            raise EasyBuildError("Failed to isolate built .whl in %s: %s", whl_paths, self.builddir)

        # Fix for https://github.com/tensorflow/tensorflow/issues/6341 on Python < 3.3
        # If the site-packages/google/__init__.py file is missing, make it an empty file.
        # This fixes the "No module named google.protobuf" error that sometimes shows up during sanity_check
        # For Python >= 3.3 the logic is reversed: The __init__.py must not exist.
        # See e.g. http://python-notes.curiousefficiency.org/en/latest/python_concepts/import_traps.html
        google_protobuf_dir = os.path.join(self.installdir, self.pylibdir, 'google', 'protobuf')
        google_init_file = os.path.join(self.installdir, self.pylibdir, 'google', '__init__.py')
        if LooseVersion(det_python_version(self.python_cmd)) < LooseVersion('3.3'):
            if os.path.isdir(google_protobuf_dir) and not is_readable(google_init_file):
                self.log.debug("Creating (empty) missing %s", google_init_file)
                write_file(google_init_file, '')
        else:
            if os.path.exists(google_init_file):
                self.log.debug("Removing %s for Python >= 3.3", google_init_file)
                remove_file(google_init_file)

        # Fix cuda header paths
        # This is needed for building custom TensorFlow ops
        if LooseVersion(self.version) < LooseVersion('1.14'):
            pyshortver = '.'.join(get_software_version('Python').split('.')[:2])
            regex_subs = [(r'#include "cuda/include/', r'#include "')]
            base_path = os.path.join(self.installdir, 'lib', 'python%s' % pyshortver, 'site-packages', 'tensorflow',
                                     'include', 'tensorflow')
            for header in glob.glob(os.path.join(base_path, 'stream_executor', 'cuda', 'cuda*.h')) + glob.glob(
                    os.path.join(base_path, 'core', 'util', 'cuda*.h')):
                apply_regex_substitutions(header, regex_subs)

    def sanity_check_step(self):
        """Custom sanity check for TensorFlow."""
        if self.python_cmd is None:
            self.prepare_python()

        custom_paths = {
            'files': ['bin/tensorboard'],
            'dirs': [self.pylibdir],
        }

        custom_commands = [
            "%s -c 'import tensorflow'" % self.python_cmd,
            # tf_should_use importsweakref.finalize, which requires backports.weakref for Python < 3.4
            "%s -c 'from tensorflow.python.util import tf_should_use'" % self.python_cmd,
        ]
        res = super(EB_TensorFlow, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)

        # test installation using MNIST tutorial examples
        if self.cfg['runtest']:
            pythonpath = os.getenv('PYTHONPATH', '')
            env.setvar('PYTHONPATH', os.pathsep.join([os.path.join(self.installdir, self.pylibdir), pythonpath]))

            mnist_pys = []

            if LooseVersion(self.version) < LooseVersion('2.0'):
                mnist_pys.append('mnist_with_summaries.py')

            if LooseVersion(self.version) < LooseVersion('1.13'):
                # mnist_softmax.py was removed in TensorFlow 1.13.x
                mnist_pys.append('mnist_softmax.py')

            for mnist_py in mnist_pys:
                datadir = tempfile.mkdtemp(suffix='-tf-%s-data' % os.path.splitext(mnist_py)[0])
                logdir = tempfile.mkdtemp(suffix='-tf-%s-logs' % os.path.splitext(mnist_py)[0])
                mnist_py = os.path.join(self.start_dir, 'tensorflow', 'examples', 'tutorials', 'mnist', mnist_py)
                cmd = "%s %s --data_dir %s --log_dir %s" % (self.python_cmd, mnist_py, datadir, logdir)
                run_cmd(cmd, log_all=True, simple=True, log_ok=True)

            # run test script (if any)
            if self.test_script:
                # copy test script to build dir before running it, to avoid that a file named 'tensorflow.py'
                # (a customized TensorFlow easyblock for example) breaks 'import tensorflow'
                test_script = os.path.join(self.builddir, os.path.basename(self.test_script))
                copy_file(self.test_script, test_script)

                run_cmd("python %s" % test_script, log_all=True, simple=True, log_ok=True)

        return res



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing HEALPix, implemented as an easyblock

@author: Kenneth Hoste (HPC-UGent)
@author: Josef Dvoracek (Institute of Physics, Czech Academy of Sciences)
"""
import os
from distutils.version import LooseVersion

import easybuild.tools.toolchain as toolchain
from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_cmd_qa
from easybuild.tools.systemtools import get_shared_lib_ext


class EB_HEALPix(ConfigureMake):
    """Support for building/installing HEALPix."""

    @staticmethod
    def extra_options():
        """There 3 variants of GCC build"""
        extra_vars = {
            'gcc_target': ['generic_gcc', "Target to use when using a GCC-based compiler toolchain", CUSTOM],
        }

        return ConfigureMake.extra_options(extra_vars)

    def __init__(self, *args, **kwargs):
        """Initialisation of custom class variables for HEALPix."""
        super(EB_HEALPix, self).__init__(*args, **kwargs)

        self.build_in_installdir = True

        # target:
        #   1: basic_gcc
        #   2: generic_gcc
        #   3: linux_icc
        #   4: optimized_gcc
        #   5: osx
        #   6: osx_icc
        self.target_string = None

        comp_fam = self.toolchain.comp_family()
        if comp_fam == toolchain.INTELCOMP:  # @UndefinedVariable
            self.target_string = 'linux_icc'
        elif comp_fam in [toolchain.DUMMY, toolchain.SYSTEM, toolchain.GCC]:  # @UndefinedVariable

            self.target_string = self.cfg['gcc_target']

            if self.target_string not in ['basic_gcc', 'generic_gcc', 'optimized_gcc']:
                raise EasyBuildError("Unknown GCC target specified: %s", self.target_string)
        else:
            raise EasyBuildError("Don't know how which C++ configuration to use for toolchain family '%s'", comp_fam)

    def extract_step(self):
        """Extract sources."""
        # strip off 'Healpix_<version>' part to avoid having everything in a subdirectory
        self.cfg['unpack_options'] = "--strip-components=1"
        super(EB_HEALPix, self).extract_step()

    def configure_step(self):
        """Custom configuration procedure for HEALPix."""

        cfitsio = get_software_root('CFITSIO')
        if not cfitsio:
            raise EasyBuildError("Failed to determine root for CFITSIO, module not loaded?")

        cmd = "./configure -L"
        qa = {
            r"Should I attempt to create these directories (Y\|n)?": 'Y',
            "full name of cfitsio library (libcfitsio.a):": '',
            r"Do you want this modification to be done (y\|N)?": 'y',
            "enter suffix for directories ():": '',
            # configure for C (2), Fortran (3), C++ (4), then exit (0)
            "Enter your choice (configuration of packages can be done in any order):": ['2', '3', '4', '0'],
        }
        std_qa = {
            r"C compiler you want to use \(\S*\):": os.environ['CC'],
            r"enter name of your F90 compiler \(\S*\):": os.environ['F90'],
            r"enter name of your C compiler \(\S*\):": os.environ['CC'],
            r"options for C compiler \([^)]*\):": os.environ['CFLAGS'],
            r"enter compilation/optimisation flags for C compiler \([^)]*\):": os.environ['CFLAGS'],
            r"compilation flags for %s compiler \([^:]*\):" % os.environ['F90']: '',
            r"enter optimisation flags for %s compiler \([^)]*\):" % os.environ['F90']: os.environ['F90FLAGS'],
            r"location of cfitsio library \(\S*\):": os.path.join(cfitsio, 'lib'),
            r"cfitsio header fitsio.h \(\S*\):": os.path.join(cfitsio, 'include'),
            r"enter command for library archiving \([^)]*\):": '',
            r"archive creation \(and indexing\) command \([^)]*\):": '',
            r"A static library is produced by default. Do you also want a shared library.*": 'y',
            r"Available configurations for C\+\+ compilation are:[\s\n\S]*" +
            r"(?P<nr>[0-9]+): %s[\s\n\S]*Choose one number:" % self.target_string: '%(nr)s',
            r"PGPLOT.[\s\n]*Do you want to enable this option \?[\s\n]*\([^)]*\) \(y\|N\)": 'N',
            r"the parallel implementation[\s\n]*Enter choice.*": '1',
            r"do you want the HEALPix/C library to include CFITSIO-related functions \? \(Y\|n\):": 'Y',
            r"\(recommended if the Healpix-F90 library is to be linked to external codes\)  \(Y\|n\):": 'Y',  # PIC -> Y
            r"Do you rather want a shared/dynamic library.*": 'n',  # shared instead static? -> N
        }
        run_cmd_qa(cmd, qa, std_qa=std_qa, log_all=True, simple=True, log_ok=True)

    def build_step(self):
        """Custom build procedure for HEALPix."""
        # disable parallel build
        self.cfg['parallel'] = '1'
        self.log.debug("Disabled parallel build")
        super(EB_HEALPix, self).build_step()

    def install_step(self):
        """No dedicated install procedure for HEALPix."""
        pass

    def make_module_extra(self):
        """additional paths"""
        txt = super(EB_HEALPix, self).make_module_extra()
        paths = {
            'CPATH': 'include',
            'LIBRARY_PATH': 'lib',
            'PATH': 'bin',
        }
        for key in sorted(paths):
            txt += self.module_generator.prepend_paths(key, os.path.join('src', 'cxx', self.target_string, paths[key]))

        return txt

    def sanity_check_step(self):
        """Custom sanity check for HEALPix."""
        binaries = [os.path.join('bin', x) for x in ['alteralm', 'anafast', 'hotspot', 'map2gif', 'median_filter',
                                                     'plmgen', 'sky_ng_sim', 'sky_ng_sim_bin', 'smoothing',
                                                     'synfast', 'ud_grade']]
        libraries = [os.path.join('lib', 'lib%s.a' % x) for x in ['chealpix', 'gif', 'healpix', 'hpxgif']]

        target_subdirs = ['bin', 'lib']
        if LooseVersion(self.version) >= LooseVersion('3'):
            # 'include' subdir is only there for recent HEALPix versions
            target_subdirs.append('include')

        custom_paths = {
            'files': binaries + libraries + [os.path.join('lib', 'libchealpix.%s' % get_shared_lib_ext())],
            'dirs': ['include'] + [os.path.join('src', 'cxx', self.target_string, x) for x in target_subdirs],
        }

        custom_commands = []
        if LooseVersion(self.version) >= LooseVersion('3'):
            custom_commands.extend([
                "cd %s && make test &> build/testresults.txt" % self.installdir,
                'test $(grep -c "test completed" %s/build/testresults.txt) -eq 4' % self.installdir,
                'test $(grep -c "normal completion" %s/build/testresults.txt) -eq 10' % self.installdir,
            ])

        super(EB_HEALPix, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)



##
# Copyright 2013-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for software using the Score-P configuration style (e.g., Cube, OTF2, Scalasca, and Score-P),
implemented as an easyblock.

@author: Kenneth Hoste (Ghent University)
@author: Bernd Mohr (Juelich Supercomputing Centre)
@author: Markus Geimer (Juelich Supercomputing Centre)
@author: Alexander Grund (TU Dresden)
"""
import easybuild.tools.toolchain as toolchain
from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.modules import get_software_root, get_software_libdir


class EB_Score_minus_P(ConfigureMake):
    """
    Support for building and installing software using the Score-P configuration style (e.g., Cube, OTF2, Scalasca,
    and Score-P).
    """

    def configure_step(self, *args, **kwargs):
        """Configure the build, set configure options for compiler, MPI and dependencies."""
        # On non-cross-compile platforms, specify compiler and MPI suite explicitly.  This is much quicker and safer
        # than autodetection.  In Score-P build-system terms, the following platforms are considered cross-compile
        # architectures:
        #
        #   - Cray XT/XE/XK/XC series
        #   - Fujitsu FX10, FX100 & K computer
        #   - IBM Blue Gene series
        #
        # Of those, only Cray is supported right now.
        tc_fam = self.toolchain.toolchain_family()
        if tc_fam != toolchain.CRAYPE:
            # --with-nocross-compiler-suite=(gcc|ibm|intel|pgi|studio)
            comp_opts = {
                # assume that system toolchain uses a system-provided GCC
                toolchain.SYSTEM: 'gcc',
                toolchain.GCC: 'gcc',
                toolchain.IBMCOMP: 'ibm',
                toolchain.INTELCOMP: 'intel',
                toolchain.NVHPC: 'pgi',
                toolchain.PGI: 'pgi',
            }
            comp_fam = self.toolchain.comp_family()
            if comp_fam in comp_opts:
                self.cfg.update('configopts', "--with-nocross-compiler-suite=%s" % comp_opts[comp_fam])
            else:
                raise EasyBuildError("Compiler family %s not supported yet (only: %s)",
                                     comp_fam, ', '.join(comp_opts.keys()))

            # --with-mpi=(bullxmpi|hp|ibmpoe|intel|intel2|intelpoe|lam|mpibull2|mpich|mpich2|mpich3|openmpi|
            #             platform|scali|sgimpt|sun)
            #
            # Notes:
            #   - intel:    Intel MPI v1.x (ancient & unsupported)
            #   - intel2:   Intel MPI v2.x and higher
            #   - intelpoe: IBM POE MPI for Intel platforms
            #   - mpich:    MPICH v1.x (ancient & unsupported)
            #   - mpich2:   MPICH2 v1.x
            #   - mpich3:   MPICH v3.x & MVAPICH2
            #               This setting actually only affects options passed to the MPI (Fortran) compiler wrappers.
            #               And since MPICH v3.x-compatible options were already supported in MVAPICH2 v1.7, it is
            #               safe to use 'mpich3' for all supported versions although MVAPICH2 is based on MPICH v3.x
            #               only since v1.9b.
            #
            # With minimal toolchains, packages using this easyblock may be built with a non-MPI toolchain (e.g., OTF2).
            # In this case, skip passing the '--with-mpi' option.
            mpi_opts = {
                toolchain.INTELMPI: 'intel2',
                toolchain.OPENMPI: 'openmpi',
                toolchain.MPICH: 'mpich3',     # In EB terms, MPICH means MPICH 3.x
                toolchain.MPICH2: 'mpich2',
                toolchain.MVAPICH2: 'mpich3',
            }
            mpi_fam = self.toolchain.mpi_family()
            if mpi_fam is not None:
                if mpi_fam in mpi_opts:
                    self.cfg.update('configopts', "--with-mpi=%s" % mpi_opts[mpi_fam])
                else:
                    raise EasyBuildError("MPI family %s not supported yet (only: %s)",
                                         mpi_fam, ', '.join(mpi_opts.keys()))

        # Auto-detection for dependencies mostly works fine, but hard specify paths anyway to have full control
        #
        # Notes:
        #   - binutils: Pass include/lib directories separately, as different directory layouts may break Score-P's
        #               configure, see https://github.com/geimer/easybuild-easyblocks/pull/4#issuecomment-219284755
        deps = {
            'binutils': ['--with-libbfd-include=%s/include',
                         '--with-libbfd-lib=%%s/%s' % get_software_libdir('binutils', fs=['libbfd.a'])],
            'libunwind': ['--with-libunwind=%s'],
            # Older versions use Cube
            'Cube': ['--with-cube=%s/bin'],
            # Recent versions of Cube are split into CubeLib and CubeW(riter)
            'CubeLib': ['--with-cubelib=%s/bin'],
            'CubeWriter': ['--with-cubew=%s/bin'],
            'CUDA': ['--enable-cuda', '--with-libcudart=%s'],
            'OTF2': ['--with-otf2=%s/bin'],
            'OPARI2': ['--with-opari2=%s/bin'],
            'PAPI': ['--with-papi-header=%s/include', '--with-papi-lib=%%s/%s' % get_software_libdir('PAPI')],
            'PDT': ['--with-pdt=%s/bin'],
            'Qt': ['--with-qt=%s'],
            'SIONlib': ['--with-sionlib=%s/bin'],
        }
        for (dep_name, dep_opts) in deps.items():
            dep_root = get_software_root(dep_name)
            if dep_root:
                for dep_opt in dep_opts:
                    try:
                        dep_opt = dep_opt % dep_root
                    except TypeError:
                        pass  # Ignore subtitution error when there is nothing to substitute
                    self.cfg.update('configopts', dep_opt)

        super(EB_Score_minus_P, self).configure_step(*args, **kwargs)



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing Gate, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
@author: Toon Willems (Ghent University)
"""
import os
import shutil
from distutils.version import LooseVersion

import easybuild.tools.toolchain as toolchain
from easybuild.easyblocks.generic.cmakemake import CMakeMake
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.run import run_cmd
from easybuild.tools.systemtools import get_shared_lib_ext


class EB_GATE(CMakeMake):
    """Support for building/installing GATE."""

    @staticmethod
    def extra_options():
        extra_vars = CMakeMake.extra_options()
        extra_vars.update({
            'default_platform': ['openPBS', "Default cluster platform to set", CUSTOM],
        })
        # Out of source build doesn't work due to sub tools beeing 'make'd
        extra_vars['separate_build_dir'][0] = False
        return extra_vars

    def __init__(self, *args, **kwargs):
        """Initialise class variables."""
        super(EB_GATE, self).__init__(*args, **kwargs)
        self.g4system = None
        self.gate_subdirs = [
            '.',
            os.path.join('cluster_tools', 'filemerger'),
            os.path.join('cluster_tools', 'jobsplitter'),
        ]

    def configure_step(self):
        """Custom configure procedure for GATE: CMake for versions 6.2 or more recent."""

        if LooseVersion(self.version) >= '6.2':
            super(EB_GATE, self).configure_step()

    def build_step(self):
        """Custom build procedure for GATE, including cluster tools."""

        if LooseVersion(self.version) < '6.2':
            # build procedure for versions older than v6.2: source env_gate.sh and run 'make'
            env_gate_script = os.path.join(self.cfg['start_dir'], 'env_gate.sh')
            self.cfg['prebuildopts'] = "source %s && %s " % (env_gate_script, self.cfg['prebuildopts'])

        if LooseVersion(self.version) <= '6.2':
            if self.toolchain.comp_family() in [toolchain.INTELCOMP]:
                # include missing library path in linker command
                self.cfg.update('buildopts', 'LD="$CXX -L$EBROOTICC/compiler/lib/intel64"')

        # redefine $CFLAGS/$CXXFLAGS via options to build command ('make')
        cflags = os.getenv('CFLAGS')
        cxxflags = "%s -DGC_DEFAULT_PLATFORM='\\\"%s\\\"'" % (os.getenv('CXXFLAGS'), self.cfg['default_platform'])
        if self.toolchain.comp_family() in [toolchain.INTELCOMP]:
            # make sure GNU macros are defined by Intel compiler
            cflags += " -gcc"
            cxxflags += " -gcc"

        # make sure right compilers and compiler options are used for building
        make_opts = {
            'CC': os.getenv('CC'),
            'CFLAGS': cflags,
            'CXX': os.getenv('CXX'),
            'CXXFLAGS': cxxflags,
            # filemerger Makefile hardcodes $LD to g++, so make sure right compiler is used for linking
            'LD': os.getenv('CXX'),
        }
        for key in sorted(make_opts):
            self.cfg.update('buildopts', '%s="%s"' % (key, make_opts[key]))

        for subdir in self.gate_subdirs:
            try:
                os.chdir(os.path.join(os.path.join(self.cfg['start_dir'], subdir)))
            except OSError as err:
                raise EasyBuildError("Failed to move to %s: %s", subdir, err)

            super(EB_GATE, self).build_step()

        try:
            os.chdir(self.cfg['start_dir'])
        except OSError as err:
            raise EasyBuildError("Failed to return to start dir %s: %s", self.cfg['start_dir'], err)

    def install_step(self):
        """Custom installation procedure for GATE."""

        if LooseVersion(self.version) >= '6.2':
            # make sure installation prefix is honored (for cluster tools, requires Makefile patch)
            self.cfg.update('installopts', 'PREFIX=%(installdir)s')

            for subdir in self.gate_subdirs:
                try:
                    os.chdir(os.path.join(os.path.join(self.cfg['start_dir'], subdir)))
                except OSError as err:
                    raise EasyBuildError("Failed to move to %s: %s", subdir, err)

                super(EB_GATE, self).install_step()
        else:
            # manually copy files for versions prior to v6.2

            # copy all the things
            try:
                shutil.rmtree(self.installdir)
                shutil.copytree(self.cfg['start_dir'], self.installdir)
            except OSError as err:
                raise EasyBuildError("Failed to copy %s to %s: %s", self.cfg['start_dir'], self.installdir, err)

            cmd = "source %s &> /dev/null && echo $G4SYSTEM" % os.path.join(self.cfg['start_dir'], 'env_gate.sh')
            out, _ = run_cmd(cmd, simple=False)
            self.g4system = out.strip()
            if self.g4system:
                self.log.debug("Value obtained for $G4SYSTEM: %s" % self.g4system)
            else:
                raise EasyBuildError("Failed to obtain value for $G4SYSTEM, not set?")

            # copy Gate libraries to 'lib' subdir in installation directory
            try:
                shlib_ext = get_shared_lib_ext()
                libdir = os.path.join(self.installdir, "lib")
                os.mkdir(libdir)
                srclibdir = os.path.join(self.cfg['start_dir'], "tmp", self.g4system, "Gate")
                for fil in os.listdir(srclibdir):
                    if fil.endswith('.%s' % shlib_ext):
                        shutil.copy2(os.path.join(srclibdir, fil), os.path.join(libdir, fil))
                        self.log.debug("Copied library %s to 'lib' install subdirectory" % fil)
            except OSError as err:
                raise EasyBuildError("Failed to copy GATE library/ies to lib directory: %s", err)

            # add link from tmp/<OS>-<comp>/gjs to lib
            js_dir = os.path.join(self.installdir, 'cluster_tools', 'jobsplitter')
            try:
                os.symlink(os.path.join(js_dir, 'tmp', self.g4system, 'gjs'), os.path.join(js_dir, 'lib'))
            except OSError as err:
                raise EasyBuildError("Failed to symlink tmp js dir to lib in %s: %s", js_dir, err)

    def make_module_extra(self):
        """Overwritten from Application to add extra txt"""
        subdir = os.getenv('G4SYSTEM', '')
        txt = super(EB_GATE, self).make_module_extra()
        path_dirs = []
        fm_bin = os.path.join('cluster_tools', 'filemerger', 'bin', subdir)
        js_dir = os.path.join('cluster_tools', 'jobsplitter')
        js_bin = os.path.join(js_dir, 'bin', subdir)
        for path_dir in [os.path.join('bin', subdir), fm_bin, js_bin]:
            if os.path.exists(os.path.join(self.installdir, path_dir)):
                path_dirs.append(path_dir)
        txt += self.module_generator.prepend_paths('PATH', path_dirs)
        txt += self.module_generator.prepend_paths('LD_LIBRARY_PATH', [os.path.join(js_dir, 'lib')])
        txt += self.module_generator.set_environment('GATEHOME', self.installdir)
        txt += self.module_generator.set_environment('GC_GATE_EXE_DIR', os.path.join(self.installdir, 'bin', subdir))
        return txt

    def sanity_check_step(self):
        """Custom sanity check for GATE."""

        if LooseVersion(self.version) >= '6.2':
            subdir = ''
            extra_files = ["bin/gjm", "bin/gjs"]
            dirs = []
            if LooseVersion(self.version) < '7.0':
                extra_files += ["Utilities/itkzlib/%s" % x for x in ['itk_zlib_mangle.h', 'zconf.h',
                                                                     'zlibDllConfig.h', 'zlib.h']]
                extra_files += ["Utilities/MetaIO/%s" % x for x in ['localMetaConfiguration.h', 'metaDTITube.h',
                                                                    'metaImage.h', 'metaMesh.h', 'metaTubeGraph.h',
                                                                    'metaUtils.h']]
        else:
            subdir = self.g4system
            extra_files = [
                os.path.join('cluster_tools', 'filemerger', 'bin', subdir, 'gjm'),
                os.path.join('cluster_tools', 'jobsplitter', 'bin', subdir, 'gjs'),
                'lib/libGate.%s' % get_shared_lib_ext(),
            ]
            dirs = ['benchmarks', 'examples']

        custom_paths = {
            'files': [os.path.join('bin', subdir, 'Gate')] + extra_files,
            'dirs': dirs,
        }
        custom_commands = ["gjs -h | grep 'This executable is compiled with %s as default'"
                           % self.cfg['default_platform']]
        super(EB_GATE, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing RepeatMasker, implemented as an easyblock
"""
from distutils.version import LooseVersion
import os

from easybuild.easyblocks.generic.tarball import Tarball
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import change_dir, patch_perl_script_autoflush
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_cmd_qa


class EB_RepeatMasker(Tarball):
    """Support for building/installing RepeatMasker."""

    def install_step(self):
        """Custom install procedure for RepeatMasker."""
        super(EB_RepeatMasker, self).install_step()

        # check for required dependencies
        perl_root = get_software_root('Perl')
        if perl_root:
            perl = os.path.join(perl_root, 'bin', 'perl')
        else:
            raise EasyBuildError("Missing required dependency: Perl")

        trf_root = get_software_root('TRF')
        if trf_root:
            trf = os.path.join(trf_root, 'trf')
        else:
            raise EasyBuildError("Missing required dependency: TRF")

        # determine which search engine to use
        # see also http://www.repeatmasker.org/RMDownload.html
        cand_search_engines = ['CrossMatch', 'RMBlast', 'WUBlast', 'HMMER']
        search_engine = None
        for dep in cand_search_engines:
            if get_software_root(dep):
                if search_engine is None:
                    search_engine = dep
                else:
                    raise EasyBuildError("Found multiple candidate search engines: %s and %s", search_engine, dep)

        if search_engine is None:
            raise EasyBuildError("No search engine found, one of these must be included as dependency: %s",
                                 ' '.join(cand_search_engines))

        change_dir(self.installdir)

        patch_perl_script_autoflush('configure')

        search_engine_bindir = os.path.join(get_software_root(search_engine), 'bin')

        if LooseVersion(self.version) >= LooseVersion('4.0.9'):
            search_engine_map = {
                'CrossMatch': '1',
                'RMBlast': '2',
                'HMMER': '3',
                'WUBlast': '4',
            }
            qa = {}
            std_qa = {
                r'\*\*TRF PROGRAM\*\*\n\n.*\n.*\n+Enter path.*': trf,
                r'.*\[ Un\-configured \]\n.*\[ Un\-configured \]\n.*\[ Un\-configured \]\n'
                r'.*\[ Un\-configured \]\n\n.*\n\n\nEnter Selection:\s*': search_engine_map[search_engine],
                r'\*\*.* INSTALLATION PATH\*\*\n\n.*\n.*\n+Enter path.*': search_engine_bindir,
                r'search engine for Repeatmasker.*': 'Y',
                r'.*\[ Configured, Default \](.*\n)*\n\nEnter Selection:\s*': '5',  # 'Done'
            }
        else:
            search_engine_map = {
                'CrossMatch': '1',
                'RMBlast': '2',
                'WUBlast': '3',
                'HMMER': '4',
            }
            qa = {
                '<PRESS ENTER TO CONTINUE>': '',
                # select search engine
                'Enter Selection:': search_engine_map[search_engine],
            }
            std_qa = {
                r'\*\*PERL PROGRAM\*\*\n([^*]*\n)+Enter path.*': perl,
                r'\*\*REPEATMASKER INSTALLATION DIRECTORY\*\*\n([^*]*\n)+Enter path.*': self.installdir,
                r'\*\*TRF PROGRAM\*\*\n([^*]*\n)+Enter path.*': trf,
                # search engine installation path (location of /bin subdirectory)
                # also enter 'Y' to confirm + '5' ("Done") to complete selection process for search engine
                r'\*\*.* INSTALLATION PATH\*\*\n([^*]*\n)+Enter path.*': search_engine_bindir + '\nY\n5',
            }

        cmd = "perl ./configure"
        run_cmd_qa(cmd, qa, std_qa=std_qa, log_all=True, simple=True, log_ok=True, maxhits=300)

    def sanity_check_step(self):
        """Custom sanity check for RepeatMasker."""

        custom_paths = {
            'files': ['RepeatMasker', 'RepeatMaskerConfig.pm'],
            'dirs': ['Libraries', 'util'],
        }

        custom_commands = ['RepeatMasker']

        super(EB_RepeatMasker, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)

    def make_module_req_guess(self):
        """Custom guesses for path-like environment variables for RepeatMaskerConfig."""
        guesses = super(EB_RepeatMasker, self).make_module_req_guess()

        guesses['PATH'] = ['']

        return guesses



##
# Copyright 2012-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBlock for installing jaxlib, implemented as an easyblock

@author: Denis Kristak (INUITS)
@author: Alexander Grund (TU Dresden)
@author: Alex Domingo (Vrije Universiteit Brussel)
"""

import os
import tempfile

from distutils.version import LooseVersion

import easybuild.tools.environment as env
from easybuild.easyblocks.generic.pythonpackage import PythonPackage
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import apply_regex_substitutions, which
from easybuild.tools.modules import get_software_root, get_software_version


class EB_jaxlib(PythonPackage):
    """Support for installing jaxlib. Extension of the existing PythonPackage easyblock"""

    @staticmethod
    def extra_options():
        """Custom easyconfig parameters specific to jaxlib."""
        extra_vars = PythonPackage.extra_options()

        extra_vars['use_pip'][0] = True
        # Run custom build script and install the generated whl file
        extra_vars['buildcmd'][0] = '%(python)s build/build.py'
        extra_vars['install_src'][0] = 'dist/*.whl'

        # Custom parameters
        extra_vars.update({
            'use_mkl_dnn': [True, "Enable support for Intel MKL-DNN", CUSTOM],
        })

        return extra_vars

    def configure_step(self):
        """Custom configure step for jaxlib."""

        super(EB_jaxlib, self).configure_step()

        binutils_root = get_software_root('binutils')
        if not binutils_root:
            raise EasyBuildError("Failed to determine installation prefix for binutils")
        config_env_vars = {
            # This is the binutils bin folder: https://github.com/tensorflow/tensorflow/issues/39263
            'GCC_HOST_COMPILER_PREFIX': os.path.join(binutils_root, 'bin'),
        }

        # Collect options for the build script
        # Used only by the build script

        # C++ flags are set through copt below
        options = ['--target_cpu_features=default']

        # Passed directly to bazel
        bazel_startup_options = [
            '--output_user_root=%s' % tempfile.mkdtemp(suffix='-bazel', dir=self.builddir),
        ]

        # Passed to the build command of bazel
        bazel_options = [
            '--jobs=%s' % self.cfg['parallel'],
            '--subcommands',
            '--action_env=PYTHONPATH',
            '--action_env=EBPYTHONPREFIXES',
        ]
        if self.toolchain.options.get('debug', None):
            bazel_options.extend([
                '--strip=never',
                '--copt="-Og"'
            ])
        # Add optimization flags set by EasyBuild each as a separate option
        bazel_options.extend(['--copt=%s' % i for i in os.environ['CXXFLAGS'].split(' ')])

        cuda_root = get_software_root('CUDA')
        if cuda_root:
            cudnn_root = get_software_root('cuDNN')
            if not cudnn_root:
                raise EasyBuildError('For CUDA-enabled builds cuDNN is also required')
            cuda_version = '.'.join(get_software_version('CUDA').split('.')[:2])  # maj.minor
            cudnn_version = '.'.join(get_software_version('cuDNN').split('.')[:3])  # maj.minor.patch
            options.extend([
                '--enable_cuda',
                '--cuda_path=' + cuda_root,
                '--cuda_compute_capabilities=' + self.cfg.get_cuda_cc_template_value('cuda_compute_capabilities'),
                '--cuda_version=' + cuda_version,
                '--cudnn_path=' + cudnn_root,
                '--cudnn_version=' + cudnn_version,
            ])

            if LooseVersion(self.version) >= LooseVersion('0.1.70'):
                nccl_root = get_software_root('NCCL')
                if nccl_root:
                    options.append('--enable_nccl')
                else:
                    options.append('--noenable_nccl')

            config_env_vars['GCC_HOST_COMPILER_PATH'] = which(os.getenv('CC'))
        else:
            options.append('--noenable_cuda')

        if self.cfg['use_mkl_dnn']:
            options.append('--enable_mkl_dnn')
        else:
            options.append('--noenable_mkl_dnn')

        # Prepend to buildopts so users can overwrite this
        self.cfg['buildopts'] = ' '.join(
            options +
            ['--bazel_startup_options="%s"' % i for i in bazel_startup_options] +
            ['--bazel_options="%s"' % i for i in bazel_options] +
            [self.cfg['buildopts']]
        )

        for key, val in sorted(config_env_vars.items()):
            env.setvar(key, val)

        # Print output of build at the end
        apply_regex_substitutions('build/build.py', [(r'  shell\(command\)', '  print(shell(command))')])



##
# Copyright 2021-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building NCCL, implemented as an easyblock

@author: Simon Branford (University of Birmingham)
"""
from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.tools.config import build_option
from easybuild.tools.systemtools import get_shared_lib_ext


class EB_NCCL(ConfigureMake):
    """Support for building NCCL."""

    def configure_step(self):
        """NCCL has no configure step"""
        pass

    def build_step(self):
        """Build NCCL"""
        # NCCL builds for all supported CUDA compute capabilities by default
        # If cuda_compute_capabilities is specified then we override this with the selected options

        # list of CUDA compute capabilities to use can be specifed in three ways (where 3 overrules 2 overrules 1):
        # (1) in the easyconfig file, via the custom cuda_compute_capabilities;
        # (2) via the EasyBuild environment variable EASYBUILD_CUDA_COMPUTE_CAPABILITIES;
        # (3) in the EasyBuild configuration, via --cuda-compute-capabilities configuration option;
        cuda_cc = build_option('cuda_compute_capabilities') or self.cfg['cuda_compute_capabilities']

        nvcc_gencode = []
        for cc in cuda_cc:
            add = cc.replace('.', '')
            nvcc_gencode.append('-gencode=arch=compute_%s,code=sm_%s' % (add, add))

        if nvcc_gencode:
            self.cfg.update('buildopts', 'NVCC_GENCODE="%s"' % ' '.join(nvcc_gencode))

        # Set PREFIX to correctly generate nccl.pc
        self.cfg.update('buildopts', "PREFIX=%s" % self.installdir)

        super(EB_NCCL, self).build_step()

    def install_step(self):
        """Install NCCL"""
        self.cfg.update('installopts', "PREFIX=%s" % self.installdir)

        super(EB_NCCL, self).install_step()

    def sanity_check_step(self):
        """Custom sanity check paths for NCCL"""
        custom_paths = {
            'files': ['include/nccl.h', 'lib/libnccl.%s' % get_shared_lib_ext(), 'lib/libnccl_static.a',
                      'lib/pkgconfig/nccl.pc'],
            'dirs': [],
        }

        super(EB_NCCL, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing HPL, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
"""

import os
import shutil

from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.run import run_cmd


class EB_HPL(ConfigureMake):
    """
    Support for building HPL (High Performance Linpack)
    - create Make.UNKNOWN
    - build with make and install
    """

    def configure_step(self, subdir=None):
        """
        Create Make.UNKNOWN file to build from
        - provide subdir argument so this can be reused in HPCC easyblock
        """

        basedir = self.cfg['start_dir']
        if subdir:
            makeincfile = os.path.join(basedir, subdir, 'Make.UNKNOWN')
            setupdir = os.path.join(basedir, subdir, 'setup')
        else:
            makeincfile = os.path.join(basedir, 'Make.UNKNOWN')
            setupdir = os.path.join(basedir, 'setup')

        try:
            os.chdir(setupdir)
        except OSError as err:
            raise EasyBuildError("Failed to change to to dir %s: %s", setupdir, err)

        cmd = "/bin/bash make_generic"

        run_cmd(cmd, log_all=True, simple=True, log_output=True)

        try:
            os.symlink(os.path.join(setupdir, 'Make.UNKNOWN'), os.path.join(makeincfile))
        except OSError as err:
            raise EasyBuildError("Failed to symlink Make.UNKNOWN from %s to %s: %s", setupdir, makeincfile, err)

        # go back
        os.chdir(self.cfg['start_dir'])

    def build_step(self):
        """
        Build with make and correct make options
        """

        for envvar in ['MPICC', 'LIBLAPACK_MT', 'CPPFLAGS', 'LDFLAGS', 'CFLAGS']:
            # environment variable may be defined but empty
            if os.getenv(envvar, None) is None:
                raise EasyBuildError("Required environment variable %s not found (no toolchain used?).", envvar)

        # build dir
        extra_makeopts = 'TOPdir="%s" ' % self.cfg['start_dir']

        # compilers
        extra_makeopts += 'CC="%(mpicc)s" MPICC="%(mpicc)s" LINKER="%(mpicc)s" ' % {'mpicc': os.getenv('MPICC')}

        # libraries: LAPACK and FFTW
        extra_makeopts += 'LAlib="%s" ' % os.getenv('LIBLAPACK_MT')

        # HPL options
        extra_makeopts += 'HPL_OPTS="%s " ' % os.getenv('CPPFLAGS')

        # linker flags
        extra_makeopts += 'LINKFLAGS="%s %s %s" ' % (os.getenv('CFLAGS'), os.getenv('LDFLAGS'), os.getenv('LIBS', ''))

        # C compilers flags
        extra_makeopts += "CCFLAGS='$(HPL_DEFS) %s' " % os.getenv('CFLAGS')

        # set options and build
        self.cfg.update('buildopts', extra_makeopts)
        super(EB_HPL, self).build_step()

    def install_step(self):
        """
        Install by copying files to install dir
        """
        srcdir = os.path.join(self.cfg['start_dir'], 'bin', 'UNKNOWN')
        destdir = os.path.join(self.installdir, 'bin')
        srcfile = None
        try:
            os.makedirs(destdir)
            for filename in ["xhpl", "HPL.dat"]:
                srcfile = os.path.join(srcdir, filename)
                shutil.copy2(srcfile, destdir)
        except OSError as err:
            raise EasyBuildError("Copying %s to installation dir %s failed: %s", srcfile, destdir, err)

    def sanity_check_step(self):
        """
        Custom sanity check for HPL
        """

        custom_paths = {
            'files': ["bin/xhpl"],
            'dirs': []
        }

        super(EB_HPL, self).sanity_check_step(custom_paths)



# #
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
# #
"""
EasyBuild support for installing the Intel Math Kernel Library (MKL), implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
@author: Ward Poelmans (Ghent University)
@author: Lumir Jasiok (IT4Innovations)
@author: Bart Oldeman (McGill University, Calcul Quebec, Compute Canada)
"""
import glob
import itertools
import os
import shutil
import tempfile
from distutils.version import LooseVersion

import easybuild.tools.environment as env
import easybuild.tools.toolchain as toolchain
from easybuild.easyblocks.generic.intelbase import IntelBase, ACTIVATION_NAME_2012, LICENSE_FILE_NAME_2012
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import apply_regex_substitutions, change_dir, mkdir, move_file, remove_dir, write_file
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_cmd
from easybuild.tools.systemtools import get_shared_lib_ext


class EB_imkl(IntelBase):
    """
    Class that can be used to install mkl
    - tested with 10.2.1.017
    -- will fail for all older versions (due to newer silent installer)
    """

    @staticmethod
    def extra_options():
        """Add easyconfig parameters custom to imkl (e.g. interfaces)."""
        extra_vars = {
            'interfaces': [True, "Indicates whether interfaces should be built", CUSTOM],
            'flexiblas': [None, "Indicates whether FlexiBLAS-compatible libraries should be built, "
                          "default from version 2021", CUSTOM],
        }
        return IntelBase.extra_options(extra_vars)

    def __init__(self, *args, **kwargs):
        """Constructor for imkl easyblock."""
        super(EB_imkl, self).__init__(*args, **kwargs)

        # make sure $MKLROOT isn't set, it's known to cause problems with the installation
        self.cfg.update('unwanted_env_vars', ['MKLROOT'])
        self.cdftlibs = []
        self.mpi_spec = None

        if LooseVersion(self.version) >= LooseVersion('2021'):
            self.mkl_basedir = os.path.join('mkl', self.version)
            if self.cfg['flexiblas'] is None:
                self.cfg['flexiblas'] = True
        else:
            self.mkl_basedir = 'mkl'
            if self.cfg['flexiblas'] is None:
                self.cfg['flexiblas'] = False

    def prepare_step(self, *args, **kwargs):
        """Prepare build environment."""

        if LooseVersion(self.version) >= LooseVersion('2017.2.174'):
            kwargs['requires_runtime_license'] = False
            super(EB_imkl, self).prepare_step(*args, **kwargs)
        else:
            super(EB_imkl, self).prepare_step(*args, **kwargs)

        # build the mkl interfaces, if desired
        if self.cfg['interfaces']:
            self.cdftlibs = ['fftw2x_cdft']
            if LooseVersion(self.version) >= LooseVersion('10.3'):
                self.cdftlibs.append('fftw3x_cdft')
            # check whether MPI_FAMILY constant is defined, so mpi_family() can be used
            if hasattr(self.toolchain, 'MPI_FAMILY') and self.toolchain.MPI_FAMILY is not None:
                mpi_spec_by_fam = {
                    toolchain.MPICH: 'mpich2',  # MPICH is MPICH v3.x, which is MPICH2 compatible
                    toolchain.MPICH2: 'mpich2',
                    toolchain.MVAPICH2: 'mpich2',
                    toolchain.OPENMPI: 'openmpi',
                }
                mpi_fam = self.toolchain.mpi_family()
                self.mpi_spec = mpi_spec_by_fam.get(mpi_fam)
                debugstr = "MPI toolchain component"
            else:
                # can't use toolchain.mpi_family, because of system toolchain
                if get_software_root('MPICH2') or get_software_root('MVAPICH2'):
                    self.mpi_spec = 'mpich2'
                elif get_software_root('OpenMPI'):
                    self.mpi_spec = 'openmpi'
                elif not get_software_root('impi'):
                    # no compatible MPI found: do not build cdft
                    self.cdftlibs = []
                debugstr = "loaded MPI module"
            if self.mpi_spec:
                self.log.debug("Determined MPI specification based on %s: %s", debugstr, self.mpi_spec)
            else:
                self.log.debug("No MPI or no compatible MPI found: do not build CDFT")

    def install_step(self):
        """
        Actual installation
        - create silent cfg file
        - execute command
        """
        silent_cfg_names_map = None
        silent_cfg_extras = None

        if LooseVersion(self.version) < LooseVersion('11.1'):
            # since imkl v11.1, silent.cfg has been slightly changed to be 'more standard'

            silent_cfg_names_map = {
                'activation_name': ACTIVATION_NAME_2012,
                'license_file_name': LICENSE_FILE_NAME_2012,
            }

        if LooseVersion(self.version) >= LooseVersion('11.1') and self.install_components is None:
            silent_cfg_extras = {
                'COMPONENTS': 'ALL',
            }

        super(EB_imkl, self).install_step(
            silent_cfg_names_map=silent_cfg_names_map,
            silent_cfg_extras=silent_cfg_extras)

    def build_mkl_fftw_interfaces(self, libdir):
        """Build the Intel MKL FFTW interfaces."""

        mkdir(libdir)

        loosever = LooseVersion(self.version)

        if loosever >= LooseVersion('10.3'):
            intsubdir = os.path.join(self.mkl_basedir, 'interfaces')
            inttarget = 'libintel64'
        else:
            intsubdir = 'interfaces'
            if self.cfg['m32']:
                inttarget = 'lib32'
            else:
                inttarget = 'libem64t'

        cmd = "make -f makefile %s" % inttarget

        # blas95 and lapack95 need more work, ignore for now
        # blas95 and lapack also need include/.mod to be processed
        fftw2libs = ['fftw2xc', 'fftw2xf']
        fftw3libs = ['fftw3xc', 'fftw3xf']

        interfacedir = os.path.join(self.installdir, intsubdir)
        change_dir(interfacedir)
        self.log.info("Changed to interfaces directory %s", interfacedir)

        compopt = None
        # determine whether we're using a non-Intel GCC-based or PGI/NVHPC-based toolchain
        # can't use toolchain.comp_family, because of system toolchain used when installing imkl
        if get_software_root('icc') or get_software_root('intel-compilers'):
            compopt = 'compiler=intel'
        elif get_software_root('PGI'):
            compopt = 'compiler=pgi'
        elif get_software_root('NVHPC'):
            compopt = 'compiler=nvhpc'
        # GCC should be the last as the above compilers also have an underlying GCC
        elif get_software_root('GCC'):
            compopt = 'compiler=gnu'
        else:
            raise EasyBuildError("Not using Intel/GCC/PGI/NVHPC compilers, "
                                 "don't know how to build wrapper libs")

        # patch makefiles for cdft wrappers when PGI or NVHPC is used as compiler
        if get_software_root('NVHPC'):
            regex_subs = [
                # nvhpc should be considered as a valid compiler
                ("intel gnu", "intel gnu nvhpc"),
                # transform 'gnu' case to 'nvhpc' case
                (r"ifeq \(\$\(compiler\),gnu\)", "ifeq ($(compiler),nvhpc)"),
                ('=gcc', '=nvc'),
            ]
        if get_software_root('PGI'):
            regex_subs = [
                # pgi should be considered as a valid compiler
                ("intel gnu", "intel gnu pgi"),
                # transform 'gnu' case to 'pgi' case
                (r"ifeq \(\$\(compiler\),gnu\)", "ifeq ($(compiler),pgi)"),
                ('=gcc', '=pgcc'),
            ]
        if get_software_root('PGI') or get_software_root('NVHPC'):
            regex_subs += [
                # correct flag to use C99 standard
                ('-std=c99', '-c99'),
                # -Wall and -Werror are not valid options for pgcc, no close equivalent
                ('-Wall', ''),
                ('-Werror', ''),
            ]
            for lib in self.cdftlibs:
                apply_regex_substitutions(os.path.join(interfacedir, lib, 'makefile'), regex_subs)

        if get_software_root('NVHPC'):
            regex_nvc_subs = [
                ('pgcc', 'nvc'),
                ('pgf95', 'nvfortran'),
                ('pgi', 'nvhpc'),
            ]
            for liball in glob.glob(os.path.join(interfacedir, '*', 'makefile')):
                apply_regex_substitutions(liball, regex_nvc_subs)

        for lib in fftw2libs + fftw3libs + self.cdftlibs:
            buildopts = [compopt]
            if lib in fftw3libs:
                buildopts.append('install_to=$INSTALL_DIR')
            elif lib in self.cdftlibs:
                if self.mpi_spec is not None:
                    buildopts.append('mpi=%s' % self.mpi_spec)

            precflags = ['']
            if lib.startswith('fftw2x') and not self.cfg['m32']:
                # build both single and double precision variants
                precflags = ['PRECISION=MKL_DOUBLE', 'PRECISION=MKL_SINGLE']

            intflags = ['']
            if lib in self.cdftlibs and not self.cfg['m32']:
                # build both 32-bit and 64-bit interfaces
                intflags = ['interface=lp64', 'interface=ilp64']

            allopts = [list(opts) for opts in itertools.product(intflags, precflags)]

            for flags, extraopts in itertools.product(['', '-fPIC'], allopts):
                tup = (lib, flags, buildopts, extraopts)
                self.log.debug("Building lib %s with: flags %s, buildopts %s, extraopts %s" % tup)

                tmpbuild = tempfile.mkdtemp(dir=self.builddir)
                self.log.debug("Created temporary directory %s" % tmpbuild)

                # always set INSTALL_DIR, SPEC_OPT, COPTS and CFLAGS
                # fftw2x(c|f): use $INSTALL_DIR, $CFLAGS and $COPTS
                # fftw3x(c|f): use $CFLAGS
                # fftw*cdft: use $INSTALL_DIR and $SPEC_OPT
                env.setvar('INSTALL_DIR', tmpbuild)
                env.setvar('SPEC_OPT', flags)
                env.setvar('COPTS', flags)
                env.setvar('CFLAGS', flags)

                intdir = os.path.join(interfacedir, lib)
                change_dir(intdir)
                self.log.info("Changed to interface %s directory %s", lib, intdir)

                fullcmd = "%s %s" % (cmd, ' '.join(buildopts + extraopts))
                res = run_cmd(fullcmd, log_all=True, simple=True)
                if not res:
                    raise EasyBuildError("Building %s (flags: %s, fullcmd: %s) failed", lib, flags, fullcmd)

                for fn in os.listdir(tmpbuild):
                    src = os.path.join(tmpbuild, fn)
                    if flags == '-fPIC':
                        # add _pic to filename
                        ff = fn.split('.')
                        fn = '.'.join(ff[:-1]) + '_pic.' + ff[-1]
                    dest = os.path.join(libdir, fn)
                    if os.path.isfile(src):
                        move_file(src, dest)
                        self.log.info("Moved %s to %s", src, dest)

                remove_dir(tmpbuild)

    def build_mkl_flexiblas(self, flexiblasdir):
        """
        Build libflexiblas_imkl_gnu_thread.so, libflexiblas_imkl_intel_thread.so,
        and libflexiblas_imkl_sequential.so. They can be used as FlexiBLAS backends
        via FLEXIBLAS_LIBRARY_PATH.
        """
        builderdir = os.path.join(self.installdir, self.mkl_basedir, 'tools', 'builder')
        change_dir(builderdir)
        mkdir(flexiblasdir, parents=True)

        # concatenate lists of all BLAS, CBLAS and LAPACK functions
        with tempfile.NamedTemporaryFile(dir=self.builddir, mode="wt", delete=False) as dst:
            listfilename = dst.name
            self.log.debug("Created temporary file %s" % listfilename)
            for lst in 'blas', 'cblas', 'lapack':
                with open(lst + "_example_list") as src:
                    shutil.copyfileobj(src, dst)

        compilerdir = os.path.join('..', '..', '..', '..', 'compiler', self.version,
                                   'linux', 'compiler', 'lib', 'intel64_lin')
        # IFACE_COMP_PART=gf gives the gfortran calling convention that FlexiBLAS expects
        cmds = ["make libintel64 IFACE_COMP_PART=gf export=%s name=%s" % (
            listfilename, os.path.join(flexiblasdir, 'libflexiblas_imkl_')) +
                s for s in ['sequential threading=sequential',
                            'gnu_thread parallel=gnu',
                            'intel_thread parallel=intel SYSTEM_LIBS="-lm -ldl -L%s"' % compilerdir]]

        for cmd in cmds:
            res = run_cmd(cmd, log_all=True, simple=True)
            if not res:
                raise EasyBuildError("Building FlexiBLAS-compatible library (cmd: %s) failed", cmd)

    def post_install_step(self):
        """
        Install group libraries and interfaces (if desired).
        """
        super(EB_imkl, self).post_install_step()

        # extract examples
        examples_subdir = os.path.join(self.installdir, self.mkl_basedir, 'examples')
        if os.path.exists(examples_subdir):
            cwd = change_dir(examples_subdir)
            for examples_tarball in glob.glob('examples_*.tgz'):
                run_cmd("tar xvzf %s -C ." % examples_tarball)
            change_dir(cwd)

        # reload the dependencies
        self.load_dependency_modules()

        shlib_ext = get_shared_lib_ext()

        if self.cfg['m32']:
            extra = {
                'libmkl.%s' % shlib_ext: 'GROUP (-lmkl_intel -lmkl_intel_thread -lmkl_core)',
                'libmkl_em64t.a': 'GROUP (libmkl_intel.a libmkl_intel_thread.a libmkl_core.a)',
                'libmkl_solver.a': 'GROUP (libmkl_solver.a)',
                'libmkl_scalapack.a': 'GROUP (libmkl_scalapack_core.a)',
                'libmkl_lapack.a': 'GROUP (libmkl_intel.a libmkl_intel_thread.a libmkl_core.a)',
                'libmkl_cdft.a': 'GROUP (libmkl_cdft_core.a)'
            }
        else:
            extra = {
                'libmkl.%s' % shlib_ext: 'GROUP (-lmkl_intel_lp64 -lmkl_intel_thread -lmkl_core)',
                'libmkl_em64t.a': 'GROUP (libmkl_intel_lp64.a libmkl_intel_thread.a libmkl_core.a)',
                'libmkl_solver.a': 'GROUP (libmkl_solver_lp64.a)',
                'libmkl_scalapack.a': 'GROUP (libmkl_scalapack_lp64.a)',
                'libmkl_lapack.a': 'GROUP (libmkl_intel_lp64.a libmkl_intel_thread.a libmkl_core.a)',
                'libmkl_cdft.a': 'GROUP (libmkl_cdft_core.a)'
            }

        loosever = LooseVersion(self.version)

        if loosever >= LooseVersion('10.3'):
            libsubdir = os.path.join(self.mkl_basedir, 'lib', 'intel64')
        else:
            if self.cfg['m32']:
                libsubdir = os.path.join('lib', '32')
            else:
                libsubdir = os.path.join('lib', 'em64t')

        libdir = os.path.join(self.installdir, libsubdir)
        for fil, txt in extra.items():
            dest = os.path.join(libdir, fil)
            if not os.path.exists(dest):
                write_file(dest, txt)

        if self.cfg['interfaces']:
            self.build_mkl_fftw_interfaces(os.path.join(self.installdir, libdir))

        if self.cfg['flexiblas']:
            self.build_mkl_flexiblas(os.path.join(self.installdir, libdir, 'flexiblas'))

    def get_mkl_fftw_interface_libs(self):
        """Returns list of library names produced by build_mkl_fftw_interfaces()"""

        if get_software_root('icc') or get_software_root('intel-compilers'):
            compsuff = '_intel'
        # check for PGI and NVHPC first, since there's a GCC underneath PGI and NVHPC too...
        elif get_software_root('PGI'):
            compsuff = '_pgi'
        elif get_software_root('NVHPC'):
            compsuff = '_nvhpc'
        elif get_software_root('GCC'):
            compsuff = '_gnu'
        else:
            raise EasyBuildError("Not using Intel/GCC/PGI/NVHPC, "
                                 "don't know compiler suffix for FFTW libraries.")

        precs = ['_double', '_single']
        ver = LooseVersion(self.version)
        if ver < LooseVersion('11'):
            # no precision suffix in libfftw2 libs before imkl v11
            precs = ['']
        fftw_vers = ['2x%s%s' % (x, prec) for x in ['c', 'f'] for prec in precs] + ['3xc', '3xf']
        pics = ['', '_pic']
        libs = ['libfftw%s%s%s.a' % (fftwver, compsuff, pic) for fftwver in fftw_vers for pic in pics]

        if self.cdftlibs:
            fftw_cdft_vers = ['2x_cdft_DOUBLE']
            if not self.cfg['m32']:
                fftw_cdft_vers.append('2x_cdft_SINGLE')
            if ver >= LooseVersion('10.3'):
                fftw_cdft_vers.append('3x_cdft')
            if ver >= LooseVersion('11.0.2'):
                bits = ['_lp64']
                if not self.cfg['m32']:
                    bits.append('_ilp64')
            else:
                # no bits suffix in cdft libs before imkl v11.0.2
                bits = ['']
            libs += ['libfftw%s%s%s.a' % x for x in itertools.product(fftw_cdft_vers, bits, pics)]

        return libs

    def sanity_check_step(self):
        """Custom sanity check paths for Intel MKL."""
        shlib_ext = get_shared_lib_ext()

        mklfiles = None
        mkldirs = None
        ver = LooseVersion(self.version)
        libs = ['libmkl_core.%s' % shlib_ext, 'libmkl_gnu_thread.%s' % shlib_ext,
                'libmkl_intel_thread.%s' % shlib_ext, 'libmkl_sequential.%s' % shlib_ext]
        extralibs = ['libmkl_blacs_intelmpi_%(suff)s.' + shlib_ext, 'libmkl_scalapack_%(suff)s.' + shlib_ext]

        if self.cfg['interfaces']:
            libs += self.get_mkl_fftw_interface_libs()

        if self.cfg['flexiblas']:
            libs += [os.path.join('flexiblas', 'libflexiblas_imkl_%s.so' % thread)
                     for thread in ['gnu_thread', 'intel_thread', 'sequential']]

        if ver >= LooseVersion('10.3') and self.cfg['m32']:
            raise EasyBuildError("Sanity check for 32-bit not implemented yet for IMKL v%s (>= 10.3)", self.version)

        if ver >= LooseVersion('10.3'):
            mkldirs = [
                os.path.join(self.mkl_basedir, 'bin'),
                os.path.join(self.mkl_basedir, 'lib', 'intel64'),
                os.path.join(self.mkl_basedir, 'include'),
            ]
            libs += [lib % {'suff': suff} for lib in extralibs for suff in ['lp64', 'ilp64']]

            mklfiles = [os.path.join(self.mkl_basedir, 'include', 'mkl.h')]
            mklfiles.extend([os.path.join(self.mkl_basedir, 'lib', 'intel64', lib) for lib in libs])

        if ver >= LooseVersion('2021'):

            mklfiles.append(os.path.join(self.mkl_basedir, 'lib', 'intel64', 'libmkl_core.%s' % shlib_ext))

        elif ver >= LooseVersion('10.3'):
            if ver < LooseVersion('11.3'):
                mkldirs.append(os.path.join(self.mkl_basedir, 'bin', 'intel64'))

            mklfiles.append(os.path.join(self.mkl_basedir, 'lib', 'intel64', 'libmkl.%s' % shlib_ext))

            if ver >= LooseVersion('10.3.4') and ver < LooseVersion('11.1'):
                mkldirs += [os.path.join('compiler', 'lib', 'intel64')]
            elif ver >= LooseVersion('2017.0.0'):
                mkldirs += [os.path.join('lib', 'intel64_lin')]
            else:
                mkldirs += [os.path.join('lib', 'intel64')]

        else:
            if self.cfg['m32']:
                lib_subdir = '32'
            else:
                lib_subdir = 'em64t'
                libs += [lib % {'suff': suff} for lib in extralibs for suff in ['lp64', 'ilp64']]

            mklfiles = [
                os.path.join('lib', lib_subdir, 'libmkl.%s' % shlib_ext),
                os.path.join('include', 'mkl.h'),
            ]
            mklfiles.extend([os.path.join('lib', lib_subdir, lib) for lib in libs])
            mkldirs = [os.path.join('lib', lib_subdir), os.path.join('include', lib_subdir), 'interfaces']

        custom_paths = {
            'files': mklfiles,
            'dirs': mkldirs,
        }

        super(EB_imkl, self).sanity_check_step(custom_paths=custom_paths)

    def make_module_req_guess(self):
        """
        A dictionary of possible directories to look for
        """
        guesses = super(EB_imkl, self).make_module_req_guess()

        if LooseVersion(self.version) >= LooseVersion('10.3'):
            if self.cfg['m32']:
                raise EasyBuildError("32-bit not supported yet for IMKL v%s (>= 10.3)", self.version)
            else:
                if LooseVersion(self.version) >= LooseVersion('2021'):
                    compiler_subdir = os.path.join('compiler', self.version, 'linux', 'compiler', 'lib', 'intel64_lin')
                    pkg_config_path = [os.path.join(self.mkl_basedir, 'tools', 'pkgconfig'),
                                       os.path.join(self.mkl_basedir, 'lib', 'pkgconfig')]
                else:
                    compiler_subdir = os.path.join('lib', 'intel64')
                    pkg_config_path = [os.path.join(self.mkl_basedir, 'bin', 'pkgconfig')]
                    guesses['MANPATH'] = ['man', os.path.join('man', 'en_US')]
                    if LooseVersion(self.version) >= LooseVersion('11.0'):
                        if LooseVersion(self.version) >= LooseVersion('11.3'):
                            guesses['MIC_LD_LIBRARY_PATH'] = [
                                os.path.join('lib', 'intel64_lin_mic'),
                                os.path.join(self.mkl_basedir, 'lib', 'mic'),
                            ]
                        elif LooseVersion(self.version) >= LooseVersion('11.1'):
                            guesses['MIC_LD_LIBRARY_PATH'] = [
                                os.path.join('lib', 'mic'),
                                os.path.join(self.mkl_basedir, 'lib', 'mic'),
                            ]
                        else:
                            guesses['MIC_LD_LIBRARY_PATH'] = [
                                os.path.join('compiler', 'lib', 'mic'),
                                os.path.join(self.mkl_basedir, 'lib', 'mic'),
                            ]
                library_path = [
                    compiler_subdir,
                    os.path.join(self.mkl_basedir, 'lib', 'intel64'),
                ]
                cpath = [
                    os.path.join(self.mkl_basedir, 'include'),
                    os.path.join(self.mkl_basedir, 'include', 'fftw'),
                ]
                guesses.update({
                    'PATH': [],
                    'LD_LIBRARY_PATH': library_path,
                    'LIBRARY_PATH': library_path,
                    'CPATH': cpath,
                    'PKG_CONFIG_PATH': pkg_config_path,
                })
                if self.cfg['flexiblas']:
                    guesses['FLEXIBLAS_LIBRARY_PATH'] = os.path.join(library_path[1], 'flexiblas')
        else:
            if self.cfg['m32']:
                guesses.update({
                    'PATH': ['bin', 'bin/ia32', 'tbb/bin/ia32'],
                    'LD_LIBRARY_PATH': ['lib', 'lib/32'],
                    'LIBRARY_PATH': ['lib', 'lib/32'],
                    'MANPATH': ['man', 'share/man', 'man/en_US'],
                })

            else:
                guesses.update({
                    'PATH': ['bin', 'bin/intel64', 'tbb/bin/em64t'],
                    'LD_LIBRARY_PATH': ['lib', 'lib/em64t'],
                    'LIBRARY_PATH': ['lib', 'lib/em64t'],
                    'MANPATH': ['man', 'share/man', 'man/en_US'],
                })
        return guesses

    def make_module_extra(self):
        """Overwritten from Application to add extra txt"""

        if 'MKL_EXAMPLES' not in self.cfg['modextravars']:
            self.cfg.update('modextravars', {
                'MKL_EXAMPLES': os.path.join(self.installdir, self.mkl_basedir, 'examples'),
            })

        txt = super(EB_imkl, self).make_module_extra()

        if LooseVersion(self.version) >= LooseVersion('2021'):
            mklroot = os.path.join(self.installdir, 'mkl', self.version)
        else:
            mklroot = os.path.join(self.installdir, 'mkl')

        txt += self.module_generator.set_environment('MKLROOT', mklroot)

        return txt



##
# Copyright 2012-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for CUDA, implemented as an easyblock

Ref: https://speakerdeck.com/ajdecon/introduction-to-the-cuda-toolkit-for-building-applications

@author: George Tsouloupas (Cyprus Institute)
@author: Fotis Georgatos (Uni.lu)
@author: Kenneth Hoste (Ghent University)
@author: Damian Alvarez (Forschungszentrum Juelich)
@author: Ward Poelmans (Free University of Brussels)
@author: Robert Mijakovic (LuxProvide S.A.)
"""
import os
import re
import stat

from distutils.version import LooseVersion

from easybuild.easyblocks.generic.binary import Binary
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.config import IGNORE
from easybuild.tools.filetools import adjust_permissions, change_dir, copy_dir, expand_glob_paths
from easybuild.tools.filetools import patch_perl_script_autoflush, remove_file, symlink, which, write_file
from easybuild.tools.run import run_cmd, run_cmd_qa
from easybuild.tools.systemtools import AARCH64, POWER, X86_64, get_cpu_architecture, get_shared_lib_ext
import easybuild.tools.environment as env

# Wrapper script definition
WRAPPER_TEMPLATE = """#!/bin/sh
echo "$@" | grep -e '-ccbin' -e '--compiler-bindir' > /dev/null
if [ $? -eq 0 ];
then
        echo "ERROR: do not set -ccbin or --compiler-bindir when using the `basename $0` wrapper"
else
        nvcc -ccbin=%s "$@"
        exit $?
fi """


class EB_CUDA(Binary):
    """
    Support for installing CUDA.
    """

    @staticmethod
    def extra_options():
        """Create a set of wrappers based on a list determined by the easyconfig file"""
        extra_vars = {
            'host_compilers': [None, "Host compilers for which a wrapper will be generated", CUSTOM]
        }
        return Binary.extra_options(extra_vars)

    def __init__(self, *args, **kwargs):
        """ Init the cuda easyblock adding a new cudaarch template var """
        myarch = get_cpu_architecture()
        if myarch == AARCH64:
            cudaarch = '_sbsa'
        elif myarch == POWER:
            cudaarch = '_ppc64le'
        elif myarch == X86_64:
            cudaarch = ''
        else:
            raise EasyBuildError("Architecture %s is not supported for CUDA on EasyBuild", myarch)

        super(EB_CUDA, self).__init__(*args, **kwargs)

        self.cfg.template_values['cudaarch'] = cudaarch
        self.cfg.generate_template_values()

    def extract_step(self):
        """Extract installer to have more control, e.g. options, patching Perl scripts, etc."""
        execpath = self.src[0]['path']
        run_cmd("/bin/sh " + execpath + " --noexec --nox11 --target " + self.builddir)
        self.src[0]['finalpath'] = self.builddir

    def install_step(self):
        """Install CUDA using Perl install script."""

        # define how to run the installer
        # script has /usr/bin/perl hardcoded, but we want to have control over which perl is being used
        if LooseVersion(self.version) <= LooseVersion("5"):
            install_interpreter = "perl"
            install_script = "install-linux.pl"
            self.cfg.update('installopts', '--prefix=%s' % self.installdir)
        elif LooseVersion(self.version) > LooseVersion("5") and LooseVersion(self.version) < LooseVersion("10.1"):
            install_interpreter = "perl"
            install_script = "cuda-installer.pl"
            # note: samples are installed by default
            self.cfg.update('installopts', "-verbose -silent -toolkitpath=%s -toolkit" % self.installdir)
        else:
            install_interpreter = ""
            install_script = "./cuda-installer"
            # samples are installed in two places with identical copies:
            # self.installdir/samples and $HOME/NVIDIA_CUDA-11.2_Samples
            # changing the second location (the one under $HOME) to a scratch location using
            # --samples --samplespath=self.builddir
            # avoids the duplicate and pollution of the home directory of the installer.
            self.cfg.update('installopts',
                            "--silent --samples --samplespath=%s --toolkit --toolkitpath=%s --defaultroot=%s" % (
                                self.builddir, self.installdir, self.installdir))
            # When eb is called via sudo -u someuser -i eb ..., the installer may try to chown samples to the
            # original user using the SUDO_USER environment variable, which fails
            if "SUDO_USER" in os.environ:
                self.log.info("SUDO_USER was defined as '%s', need to unset it to avoid problems..." %
                              os.environ["SUDO_USER"])
                del os.environ["SUDO_USER"]

        if LooseVersion("10.0") < LooseVersion(self.version) < LooseVersion("10.2") and get_cpu_architecture() == POWER:
            # Workaround for
            # https://devtalk.nvidia.com/default/topic/1063995/cuda-setup-and-installation/cuda-10-1-243-10-1-update-2-ppc64le-run-file-installation-issue/
            install_script = " && ".join([
                "mkdir -p %(installdir)s/targets/ppc64le-linux/include",
                "([ -e %(installdir)s/include ] || ln -s targets/ppc64le-linux/include %(installdir)s/include)",
                "cp -r %(builddir)s/builds/cublas/src %(installdir)s/.",
                install_script
            ]) % {
                'installdir': self.installdir,
                'builddir': self.builddir
            }

        # Use C locale to avoid localized questions and crash on CUDA 10.1
        self.cfg.update('preinstallopts', "export LANG=C && ")

        cmd = "%(preinstallopts)s %(interpreter)s %(script)s %(installopts)s" % {
            'preinstallopts': self.cfg['preinstallopts'],
            'interpreter': install_interpreter,
            'script': install_script,
            'installopts': self.cfg['installopts']
        }

        # prepare for running install script autonomously
        qanda = {}
        stdqa = {
            # this question is only asked if CUDA tools are already available system-wide
            r"Would you like to remove all CUDA files under .*? (yes/no/abort): ": "no",
        }
        noqanda = [
            r"^Configuring",
            r"Installation Complete",
            r"Verifying archive integrity.*",
            r"^Uncompressing NVIDIA CUDA",
            r".* -> .*",
        ]

        # patch install script to handle Q&A autonomously
        if install_interpreter == "perl":
            patch_perl_script_autoflush(os.path.join(self.builddir, install_script))
            p5lib = os.getenv('PERL5LIB', '')
            if p5lib == '':
                p5lib = self.builddir
            else:
                p5lib = os.pathsep.join([self.builddir, p5lib])
            env.setvar('PERL5LIB', p5lib)

        # make sure $DISPLAY is not defined, which may lead to (weird) problems
        # this is workaround for not being able to specify --nox11 to the Perl install scripts
        if 'DISPLAY' in os.environ:
            os.environ.pop('DISPLAY')

        # cuda-installer creates /tmp/cuda-installer.log (ignoring TMPDIR)
        # Try to remove it before running the installer.
        # This will fail with a usable error if it can't be removed
        # instead of segfaulting in the cuda-installer.
        remove_file('/tmp/cuda-installer.log')

        # overriding maxhits default value to 1000 (seconds to wait for nothing to change in the output
        # without seeing a known question)
        run_cmd_qa(cmd, qanda, std_qa=stdqa, no_qa=noqanda, log_all=True, simple=True, maxhits=1000)

        # Remove the cuda-installer log file
        remove_file('/tmp/cuda-installer.log')

        # check if there are patches to apply
        if len(self.src) > 1:
            for patch in self.src[1:]:
                self.log.debug("Running patch %s", patch['name'])
                run_cmd("/bin/sh " + patch['path'] + " --accept-eula --silent --installdir=" + self.installdir)

    def post_install_step(self):
        """
        Create wrappers for the specified host compilers, generate the appropriate stub symlinks,
        and create version independent pkgconfig files
        """
        def create_wrapper(wrapper_name, wrapper_comp):
            """Create for a particular compiler, with a particular name"""
            wrapper_f = os.path.join(self.installdir, 'bin', wrapper_name)
            write_file(wrapper_f, WRAPPER_TEMPLATE % wrapper_comp)
            perms = stat.S_IXUSR | stat.S_IRUSR | stat.S_IXGRP | stat.S_IRGRP | stat.S_IXOTH | stat.S_IROTH
            adjust_permissions(wrapper_f, perms)

        # Prepare wrappers to handle a default host compiler other than g++
        for comp in (self.cfg['host_compilers'] or []):
            create_wrapper('nvcc_%s' % comp, comp)

        ldconfig = which('ldconfig', log_ok=False, on_error=IGNORE)
        sbin_dirs = ['/sbin', '/usr/sbin']
        if not ldconfig:
            # ldconfig is usually in /sbin or /usr/sbin
            for cand_path in sbin_dirs:
                if os.path.exists(os.path.join(cand_path, 'ldconfig')):
                    ldconfig = os.path.join(cand_path, 'ldconfig')
                    break

        # fail if we couldn't find ldconfig, because it's really needed
        if ldconfig:
            self.log.info("ldconfig found at %s", ldconfig)
        else:
            path = os.environ.get('PATH', '')
            raise EasyBuildError("Unable to find 'ldconfig' in $PATH (%s), nor in any of %s", path, sbin_dirs)

        stubs_dir = os.path.join(self.installdir, 'lib64', 'stubs')
        # Run ldconfig to create missing symlinks in the stubs directory (libcuda.so.1, etc)
        cmd = ' '.join([ldconfig, '-N', stubs_dir])
        run_cmd(cmd)

        # GCC searches paths in LIBRARY_PATH and the system paths suffixed with ../lib64 or ../lib first
        # This means stubs/../lib64 is searched before the system /lib64 folder containing a potentially older libcuda.
        # See e.g. https://github.com/easybuilders/easybuild-easyconfigs/issues/12348
        # Workaround: Create a copy that matches this pattern
        new_stubs_dir = os.path.join(self.installdir, 'stubs')
        copy_dir(stubs_dir, os.path.join(new_stubs_dir, 'lib64'))
        # Also create the lib dir as a symlink
        symlink('lib64', os.path.join(new_stubs_dir, 'lib'), use_abspath_source=False)

        # Packages like xpra look for version independent pc files.
        # See e.g. https://github.com/Xpra-org/xpra/blob/master/setup.py#L206
        # Distros provide these files, so let's do it here too
        pkgconfig_dir = os.path.join(self.installdir, 'pkgconfig')
        if os.path.exists(pkgconfig_dir):
            pc_files = expand_glob_paths([os.path.join(pkgconfig_dir, '*.pc')])
            cwd = change_dir(pkgconfig_dir)
            for pc_file in pc_files:
                pc_file = os.path.basename(pc_file)
                link = re.sub('-[0-9]*.?[0-9]*(.[0-9]*)?.pc', '.pc', pc_file)
                symlink(pc_file, link, use_abspath_source=False)
            change_dir(cwd)

        super(EB_CUDA, self).post_install_step()

    def sanity_check_step(self):
        """Custom sanity check for CUDA."""

        shlib_ext = get_shared_lib_ext()

        chk_libdir = ["lib64", "lib"]
        culibs = ["cublas", "cudart", "cufft", "curand", "cusparse"]
        custom_paths = {
            'files': [os.path.join("bin", x) for x in ["fatbinary", "nvcc", "nvlink", "ptxas"]] +
            [os.path.join("%s", "lib%s.%s") % (x, y, shlib_ext) for x in chk_libdir for y in culibs],
            'dirs': ["include"],
        }

        # Samples moved to https://github.com/nvidia/cuda-samples
        if LooseVersion(self.version) > LooseVersion('5') and LooseVersion(self.version) < LooseVersion('11.6'):
            custom_paths['files'].append(os.path.join('samples', 'Makefile'))
        if LooseVersion(self.version) < LooseVersion('7'):
            custom_paths['files'].append(os.path.join('open64', 'bin', 'nvopencc'))
        if LooseVersion(self.version) >= LooseVersion('7'):
            custom_paths['files'].append(os.path.join("extras", "CUPTI", "lib64", "libcupti.%s") % shlib_ext)
            custom_paths['dirs'].append(os.path.join("extras", "CUPTI", "include"))

        # Just a subset of files are checked, since the whole list is likely to change,
        # and irrelevant in most cases anyway
        if os.path.exists(os.path.join(self.installdir, 'pkgconfig')):
            pc_files = ['cublas.pc', 'cudart.pc', 'cuda.pc']
            custom_paths['files'].extend(os.path.join('pkgconfig', x) for x in pc_files)

        super(EB_CUDA, self).sanity_check_step(custom_paths=custom_paths)

    def make_module_extra(self):
        """Set the install directory as CUDA_HOME, CUDA_ROOT, CUDA_PATH."""

        # avoid adding of installation directory to $PATH (cfr. Binary easyblock) since that may cause trouble,
        # for example when there's a clash between command name and a subdirectory in the installation directory
        # (like compute-sanitizer)
        self.cfg['prepend_to_path'] = False

        txt = super(EB_CUDA, self).make_module_extra()
        txt += self.module_generator.set_environment('CUDA_HOME', self.installdir)
        txt += self.module_generator.set_environment('CUDA_ROOT', self.installdir)
        txt += self.module_generator.set_environment('CUDA_PATH', self.installdir)
        self.log.debug("make_module_extra added this: %s", txt)
        return txt

    def make_module_req_guess(self):
        """Specify CUDA custom values for PATH etc."""

        guesses = super(EB_CUDA, self).make_module_req_guess()

        # The dirs should be in the order ['open64/bin', 'bin']
        bin_path = []
        if LooseVersion(self.version) < LooseVersion('7'):
            bin_path.append(os.path.join('open64', 'bin'))
        bin_path.append('bin')

        lib_path = ['lib64']
        inc_path = ['include']
        if LooseVersion(self.version) >= LooseVersion('7'):
            lib_path.append(os.path.join('extras', 'CUPTI', 'lib64'))
            inc_path.append(os.path.join('extras', 'CUPTI', 'include'))
            bin_path.append(os.path.join('nvvm', 'bin'))
            lib_path.append(os.path.join('nvvm', 'lib64'))
            inc_path.append(os.path.join('nvvm', 'include'))

        guesses.update({
            'CPATH': inc_path,
            'LD_LIBRARY_PATH': lib_path,
            'LIBRARY_PATH': ['lib64', os.path.join('stubs', 'lib64')],
            'PATH': bin_path,
            'PKG_CONFIG_PATH': ['pkgconfig'],
        })

        return guesses



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for installing only the MPI interfaces for FFTW,
implemented as an easyblock

@author: Bart Oldeman (McGill University, Calcul Quebec, Digital Research Alliance of Canada)
"""
import os
import glob

from easybuild.easyblocks.fftw import EB_FFTW
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.modules import get_software_root
from easybuild.tools.filetools import remove


class EB_FFTW_period_MPI(EB_FFTW):
    """Support for building/installing FFTW.MPI"""

    @staticmethod
    def extra_options():
        """Modify defaults for custom easyconfig parameters for FFTW."""

        extra_vars = EB_FFTW.extra_options()
        # change defaults for unneeded or impossible options for MPI libraries
        extra_vars['with_openmp'][0] = False
        extra_vars['with_threads'][0] = False
        extra_vars['with_quad_prec'][0] = False
        return extra_vars

    def prepare_step(self, *args, **kwargs):
        """Custom prepare step: make sure FFTW is available as dependency."""
        super(EB_FFTW_period_MPI, self).prepare_step(*args, **kwargs)

        fftw_root = get_software_root('FFTW')
        if not fftw_root:
            raise EasyBuildError("Required FFTW dependency is missing!")

    def post_install_step(self):
        """Custom post install step for FFTW.MPI"""

        # remove everything except include files that are already in non-MPI FFTW dependency.
        remove(glob.glob(os.path.join(self.installdir, 'lib*', 'libfftw.*')) +
               glob.glob(os.path.join(self.installdir, 'lib*', 'libfftw[lf].*')) +
               glob.glob(os.path.join(self.installdir, 'lib*/pkgconfig')) +
               glob.glob(os.path.join(self.installdir, 'lib*/cmake')) +
               [os.path.join(self.installdir, p) for p in ['bin', 'share']])
        super(EB_FFTW_period_MPI, self).post_install_step()

    def sanity_check_step(self):
        """Custom sanity check for FFTW.MPI: check if all libraries/headers for MPI interfaces are there."""
        super(EB_FFTW_period_MPI, self).sanity_check_step(mpionly=True)
